import logging

from logging.handlers import TimedRotatingFileHandler
from typing import Dict, Optional

from metricflow.errors.errors import SqlClientCreationException, MetricFlowInitException
from metricflow.configuration.config_handler import ConfigHandler
from metricflow.configuration.constants import CONFIG_DBT_PROFILE, CONFIG_DBT_REPO, CONFIG_DBT_TARGET, CONFIG_DWH_SCHEMA
from metricflow.engine.metricflow_engine import MetricFlowEngine
from metricflow.engine.utils import build_user_configured_model_from_config, build_user_configured_model_from_dbt_config
from metricflow.model.semantic_model import SemanticModel
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql_clients.sql_utils import make_sql_client_from_config
from metricflow.model.objects.user_configured_model import UserConfiguredModel

logger = logging.getLogger(__name__)


class CLIContext:
    """Context for MetricFlow CLI."""

    def __init__(self) -> None:  # noqa: D
        self.verbose = False
        self._mf: Optional[MetricFlowEngine] = None
        self._sql_client: Optional[SqlClient] = None
        self._user_configured_model: Optional[UserConfiguredModel] = None
        self._semantic_model: Optional[SemanticModel] = None
        self._mf_system_schema: Optional[str] = None
        self._model_path_is_for_dbt: Optional[bool] = None
        self.config = ConfigHandler()
        self._configure_logging()

    def _configure_logging(self) -> None:  # noqa: D
        log_format = "%(asctime)s %(levelname)s %(filename)s:%(lineno)d [%(threadName)s] - %(message)s"
        logging.basicConfig(level=logging.INFO, format=log_format)

        log_file_handler = TimedRotatingFileHandler(
            filename=self.config.log_file_path,
            # Rotate every day to a new file, keep 7 days worth.
            when="D",
            interval=1,
            backupCount=7,
        )

        formatter = logging.Formatter(fmt=log_format)
        log_file_handler.setFormatter(formatter)
        log_file_handler.setLevel(logging.INFO)

        root_logger = logging.getLogger()
        # StreamHandler to the console would have been setup by logging.basicConfig
        for handler in root_logger.handlers:
            handler.setLevel(logging.CRITICAL)
        root_logger.addHandler(log_file_handler)

    @property
    def mf_system_schema(self) -> str:  # noqa: D
        if self._mf_system_schema is None:
            self._mf_system_schema = self.config.get_value(CONFIG_DWH_SCHEMA)
        assert self._mf_system_schema
        return self._mf_system_schema

    def __initialize_sql_client(self) -> None:
        """Initializes the SqlClient given the credentials."""
        try:
            self._sql_client = make_sql_client_from_config(self.config)
        except Exception as e:
            raise SqlClientCreationException from e

    @property
    def sql_client(self) -> SqlClient:  # noqa: D
        if self._sql_client is None:
            # Initialize the SqlClient if not set
            self.__initialize_sql_client()
        assert self._sql_client is not None
        return self._sql_client

    def run_health_checks(self) -> Dict[str, Dict[str, str]]:
        """Execute the DB health checks."""
        try:
            return self.sql_client.health_checks(self.mf_system_schema)
        except Exception as e:
            raise SqlClientCreationException from e

    def _initialize_metricflow_engine(self) -> None:
        """Initialize the MetricFlowEngine."""
        try:
            self._mf = MetricFlowEngine.from_config(self.config)
        except Exception as e:
            raise MetricFlowInitException from e

    @property
    def mf(self) -> MetricFlowEngine:  # noqa: D
        if self._mf is None:
            self._initialize_metricflow_engine()
        assert self._mf is not None
        return self._mf

    def _build_semantic_model(self) -> None:
        """Get the path to the models and create a corresponding SemanticModel."""
        self._semantic_model = SemanticModel(self.user_configured_model)

    @property
    def semantic_model(self) -> SemanticModel:  # noqa: D
        if self._semantic_model is None:
            self._build_semantic_model()
        assert self._semantic_model is not None
        return self._semantic_model

    @property
    def model_path_is_for_dbt(self) -> bool:  # noqa: D
        if self._model_path_is_for_dbt is None:
            config_value = self.config.get_value(key=CONFIG_DBT_REPO) or ""
            self._model_path_is_for_dbt = config_value.lower() in ["yes", "y", "true", "t", "1"]

        return self._model_path_is_for_dbt

    @property
    def user_configured_model(self) -> UserConfiguredModel:  # noqa: D
        if self._user_configured_model is None:
            if self.model_path_is_for_dbt:
                dbt_profile = self.config.get_value(CONFIG_DBT_PROFILE)
                dbt_target = self.config.get_value(CONFIG_DBT_TARGET)

                self._user_configured_model = build_user_configured_model_from_dbt_config(
                    handler=self.config, profile=dbt_profile, target=dbt_target
                )
            else:
                self._user_configured_model = build_user_configured_model_from_config(self.config)

        assert self._user_configured_model is not None
        return self._user_configured_model



from collections import defaultdict
from typing import DefaultDict, Dict, List, Set

from more_itertools import bucket

from metricflow.aggregation_properties import AggregationType
from metricflow.instances import MetricModelReference
from metricflow.model.objects.elements.dimension import DimensionType
from metricflow.model.objects.metric import Metric
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.validations.unique_valid_name import UniqueAndValidNameRule
from metricflow.model.validations.validator_helpers import (
    DataSourceElementContext,
    DataSourceElementReference,
    DataSourceElementType,
    FileContext,
    MetricContext,
    ModelValidationRule,
    ValidationIssueType,
    ValidationError,
    ValidationWarning,
    validate_safely,
)
from metricflow.references import MeasureReference


class DataSourceMeasuresUniqueRule(ModelValidationRule):
    """Asserts all measure names are unique across the model."""

    @staticmethod
    @validate_safely(
        whats_being_done="running model validation ensuring measures exist in only one configured data source"
    )
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []

        measure_references_to_data_sources: Dict[MeasureReference, List] = defaultdict(list)
        for data_source in model.data_sources:
            for measure in data_source.measures:
                if measure.reference in measure_references_to_data_sources:
                    issues.append(
                        ValidationError(
                            context=DataSourceElementContext(
                                file_context=FileContext.from_metadata(metadata=data_source.metadata),
                                data_source_element=DataSourceElementReference(
                                    data_source_name=data_source.name, element_name=measure.name
                                ),
                                element_type=DataSourceElementType.MEASURE,
                            ),
                            message=f"Found measure with name {measure.name} in multiple data sources with names "
                            f"({measure_references_to_data_sources[measure.reference]})",
                        )
                    )
                measure_references_to_data_sources[measure.reference].append(data_source.name)

        return issues


class MeasureConstraintAliasesRule(ModelValidationRule):
    """Checks that aliases are configured correctly for constrained measure references

    These are, currently, only applicable for Metric types, since the MetricInputMeasure is only
    """

    @staticmethod
    @validate_safely(whats_being_done="ensuring measures aliases are set when required")
    def _validate_required_aliases_are_set(metric: Metric, metric_context: MetricContext) -> List[ValidationIssueType]:
        """Checks if valid aliases are set on the input measure references where they are required

        Aliases are required whenever there are 2 or more input measures with the same measure
        reference with different constraints. When this happens, we require aliases for all
        constrained measures for the sake of clarity. Any unconstrained measure does not
        need an alias, since it always relies on the original measure specification.

        At this time aliases are required for ratio metrics, but eventually we could relax that requirement
        if we can find an automatic aliasing scheme for numerator/denominator that we feel comfortable using.
        """
        issues: List[ValidationIssueType] = []

        if len(metric.measure_references) == len(set(metric.measure_references)):
            # All measure references are unique, so disambiguation via aliasing is not necessary
            return issues

        # Note: more_itertools.bucket does not produce empty groups
        input_measures_by_name = bucket(metric.input_measures, lambda x: x.name)
        for name in input_measures_by_name:
            input_measures = list(input_measures_by_name[name])

            if len(input_measures) == 1:
                continue

            distinct_input_measures = set(input_measures)
            if len(distinct_input_measures) == 1:
                # Warn whenever multiple identical references exist - we will consolidate these but it might be
                # a meaningful oversight if constraints and aliases are specified
                issues.append(
                    ValidationWarning(
                        context=metric_context,
                        message=(
                            f"Metric {metric.name} has multiple identical input measures specifications for measure "
                            f"{name}. This might be hiding a semantic error. Input measure specification: "
                            f"{input_measures[0]}."
                        ),
                    )
                )
                continue

            constrained_measures_without_aliases = [
                measure for measure in input_measures if measure.constraint is not None and measure.alias is None
            ]
            if constrained_measures_without_aliases:
                issues.append(
                    ValidationError(
                        context=metric_context,
                        message=(
                            f"Metric {metric.name} depends on multiple different constrained versions of measure "
                            f"{name}. In such cases, aliases must be provided, but the following input measures have "
                            f"constraints specified without an alias: {constrained_measures_without_aliases}."
                        ),
                    )
                )

        return issues

    @staticmethod
    @validate_safely(whats_being_done="checking constrained measures are aliased properly")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:
        """Ensures measures that might need an alias have one set, and that the alias is distinct

        We do not allow aliases to collide with other alias or measure names, since that could create
        ambiguity at query time or cause issues if users ever restructure their models.
        """
        issues: List[ValidationIssueType] = []

        measure_names = _get_measure_names_from_model(model)
        measure_alias_to_metrics: DefaultDict[str, List[str]] = defaultdict(list)
        for metric in model.metrics:
            metric_context = MetricContext(
                file_context=FileContext.from_metadata(metadata=metric.metadata),
                metric=MetricModelReference(metric_name=metric.name),
            )

            issues += MeasureConstraintAliasesRule._validate_required_aliases_are_set(
                metric=metric, metric_context=metric_context
            )

            aliased_measures = [
                input_measure for input_measure in metric.input_measures if input_measure.alias is not None
            ]

            for measure in aliased_measures:
                assert measure.alias, "Type refinement assertion, previous filter should ensure this is true"
                issues += UniqueAndValidNameRule.check_valid_name(measure.alias, metric_context)
                if measure.alias in measure_names:
                    issues.append(
                        ValidationError(
                            context=metric_context,
                            message=(
                                f"Alias `{measure.alias}` for measure `{measure.name}` conflicts with measure names "
                                f"defined elsewhere in the model! This can cause ambiguity for certain types of "
                                f"query. Please choose another alias."
                            ),
                        )
                    )
                if measure.alias in measure_alias_to_metrics:
                    issues.append(
                        ValidationError(
                            context=metric_context,
                            message=(
                                f"Measure alias {measure.alias} conflicts with a measure alias used elsewhere in the "
                                f"model! This can cause ambiguity for certain types of query. Please choose another "
                                f"alias, or, if the measures are constrained in the same way, consider centralizing "
                                f"that definition in a new data source. Measure specification: {measure}. Existing "
                                f"metrics with that measure alias used: {measure_alias_to_metrics[measure.alias]}"
                            ),
                        )
                    )

                measure_alias_to_metrics[measure.alias].append(metric.name)

        return issues


class MetricMeasuresRule(ModelValidationRule):
    """Checks that the measures referenced in the metrics exist."""

    @staticmethod
    @validate_safely(whats_being_done="checking all measures referenced by the metric exist")
    def _validate_metric_measure_references(metric: Metric, valid_measure_names: Set[str]) -> List[ValidationIssueType]:
        issues: List[ValidationIssueType] = []

        for measure_reference in metric.measure_references:
            if measure_reference.element_name not in valid_measure_names:
                issues.append(
                    ValidationError(
                        context=MetricContext(
                            file_context=FileContext.from_metadata(metadata=metric.metadata),
                            metric=MetricModelReference(metric_name=metric.name),
                        ),
                        message=(
                            f"Measure {measure_reference.element_name} referenced in metric {metric.name} is not "
                            f"defined in the model!"
                        ),
                    )
                )
        return issues

    @staticmethod
    @validate_safely(whats_being_done="running model validation ensuring metric measures exist")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []
        valid_measure_names = _get_measure_names_from_model(model)

        for metric in model.metrics or []:
            issues += MetricMeasuresRule._validate_metric_measure_references(
                metric=metric, valid_measure_names=valid_measure_names
            )
        return issues


class MeasuresNonAdditiveDimensionRule(ModelValidationRule):
    """Checks that the measure's non_additive_dimensions are properly defined."""

    @staticmethod
    @validate_safely(whats_being_done="ensuring that a measure's non_additive_dimensions is valid")
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []
        for data_source in model.data_sources or []:
            for measure in data_source.measures:
                non_additive_dimension = measure.non_additive_dimension
                if non_additive_dimension is None:
                    continue
                agg_time_dimension = next(
                    (
                        dim
                        for dim in data_source.dimensions
                        if measure.checked_agg_time_dimension.element_name == dim.name
                    ),
                    None,
                )
                if agg_time_dimension is None:
                    # Sanity check, should never hit this
                    issues.append(
                        ValidationError(
                            context=DataSourceElementContext(
                                file_context=FileContext.from_metadata(metadata=data_source.metadata),
                                data_source_element=DataSourceElementReference(
                                    data_source_name=data_source.name, element_name=measure.name
                                ),
                                element_type=DataSourceElementType.MEASURE,
                            ),
                            message=(
                                f"Measure '{measure.name}' has a agg_time_dimension of {measure.checked_agg_time_dimension.element_name} "
                                f"that is not defined as a dimension in data source '{data_source.name}'."
                            ),
                        )
                    )
                    continue

                # Validates that the non_additive_dimension exists as a time dimension in the data source
                matching_dimension = next(
                    (dim for dim in data_source.dimensions if non_additive_dimension.name == dim.name), None
                )
                if matching_dimension is None:
                    issues.append(
                        ValidationError(
                            context=DataSourceElementContext(
                                file_context=FileContext.from_metadata(metadata=data_source.metadata),
                                data_source_element=DataSourceElementReference(
                                    data_source_name=data_source.name, element_name=measure.name
                                ),
                                element_type=DataSourceElementType.MEASURE,
                            ),
                            message=(
                                f"Measure '{measure.name}' has a non_additive_dimension with name '{non_additive_dimension.name}' "
                                f"that is not defined as a dimension in data source '{data_source.name}'."
                            ),
                        )
                    )
                if matching_dimension:
                    # Check that it's a time dimension
                    if matching_dimension.type != DimensionType.TIME:
                        issues.append(
                            ValidationError(
                                context=DataSourceElementContext(
                                    file_context=FileContext.from_metadata(metadata=data_source.metadata),
                                    data_source_element=DataSourceElementReference(
                                        data_source_name=data_source.name, element_name=measure.name
                                    ),
                                    element_type=DataSourceElementType.MEASURE,
                                ),
                                message=(
                                    f"Measure '{measure.name}' has a non_additive_dimension with name '{non_additive_dimension.name}' "
                                    f"that is defined as a categorical dimension which is not supported."
                                ),
                            )
                        )

                    # Validates that the non_additive_dimension time_granularity is >= agg_time_dimension time_granularity
                    if (
                        matching_dimension.type_params
                        and agg_time_dimension.type_params
                        and (
                            matching_dimension.type_params.time_granularity
                            != agg_time_dimension.type_params.time_granularity
                        )
                    ):
                        issues.append(
                            ValidationError(
                                context=DataSourceElementContext(
                                    file_context=FileContext.from_metadata(metadata=data_source.metadata),
                                    data_source_element=DataSourceElementReference(
                                        data_source_name=data_source.name, element_name=measure.name
                                    ),
                                    element_type=DataSourceElementType.MEASURE,
                                ),
                                message=(
                                    f"Measure '{measure.name}' has a non_additive_dimension with name '{non_additive_dimension.name}' that has "
                                    f"a base time granularity ({matching_dimension.type_params.time_granularity.name}) that is not equal to the measure's "
                                    f"agg_time_dimension {agg_time_dimension.name} with a base granularity of ({agg_time_dimension.type_params.time_granularity.name})."
                                ),
                            )
                        )

                # Validates that the window_choice is either MIN/MAX
                if non_additive_dimension.window_choice not in {AggregationType.MIN, AggregationType.MAX}:
                    issues.append(
                        ValidationError(
                            context=DataSourceElementContext(
                                file_context=FileContext.from_metadata(metadata=data_source.metadata),
                                data_source_element=DataSourceElementReference(
                                    data_source_name=data_source.name, element_name=measure.name
                                ),
                                element_type=DataSourceElementType.MEASURE,
                            ),
                            message=(
                                f"Measure '{measure.name}' has a non_additive_dimension with an invalid 'window_choice' of '{non_additive_dimension.window_choice.value}'. "
                                f"Only choices supported are 'min' or 'max'."
                            ),
                        )
                    )

                # Validates that all window_groupings are identifiers
                identifiers_in_data_source = {identifier.name for identifier in data_source.identifiers}
                window_groupings = set(non_additive_dimension.window_groupings)
                intersected_identifiers = window_groupings.intersection(identifiers_in_data_source)
                if len(intersected_identifiers) != len(window_groupings):
                    issues.append(
                        ValidationError(
                            context=DataSourceElementContext(
                                file_context=FileContext.from_metadata(metadata=data_source.metadata),
                                data_source_element=DataSourceElementReference(
                                    data_source_name=data_source.name, element_name=measure.name
                                ),
                                element_type=DataSourceElementType.MEASURE,
                            ),
                            message=(
                                f"Measure '{measure.name}' has a non_additive_dimension with an invalid 'window_groupings'. "
                                f"These identifiers {window_groupings.difference(intersected_identifiers)} do not exist in the data source."
                            ),
                        )
                    )

        return issues


class CountAggregationExprRule(ModelValidationRule):
    """Checks that COUNT measures have an expr provided."""

    @staticmethod
    @validate_safely(
        whats_being_done="running model validation ensuring expr exist for measures with count aggregation"
    )
    def validate_model(model: UserConfiguredModel) -> List[ValidationIssueType]:  # noqa: D
        issues: List[ValidationIssueType] = []

        for data_source in model.data_sources:
            for measure in data_source.measures:
                context = DataSourceElementContext(
                    file_context=FileContext.from_metadata(metadata=data_source.metadata),
                    data_source_element=DataSourceElementReference(
                        data_source_name=data_source.name, element_name=measure.name
                    ),
                    element_type=DataSourceElementType.MEASURE,
                )
                if measure.agg == AggregationType.COUNT and measure.expr is None:
                    issues.append(
                        ValidationError(
                            context=context,
                            message=(
                                f"Measure '{measure.name}' uses a COUNT aggregation, which requires an expr to be provided. "
                                f"Provide 'expr: 1' if a count of all rows is desired."
                            ),
                        )
                    )
                if (
                    measure.agg == AggregationType.COUNT
                    and measure.expr
                    and measure.expr.lower().startswith("distinct ")
                ):
                    # TODO: Expand this to include SUM and potentially AVG agg types as well
                    # Note expansion of this guard requires the addition of sum_distinct and avg_distinct agg types
                    # or else an adjustment to the error message below.
                    issues.append(
                        ValidationError(
                            context=context,
                            message=(
                                f"Measure '{measure.name}' uses a '{measure.agg.value}' aggregation with a DISTINCT expr: "
                                f"'{measure.expr}. This is not supported, as it effectively converts an additive "
                                f"measure into a non-additive one, and this could cause certain queries to return "
                                f"incorrect results. Please use the {measure.agg.value}_distinct aggregation type."
                            ),
                        )
                    )
        return issues


def _get_measure_names_from_model(model: UserConfiguredModel) -> Set[str]:
    """Return every distinct measure name specified in the model"""
    measure_names = set()
    for data_source in model.data_sources:
        for measure in data_source.measures:
            measure_names.add(measure.reference.element_name)

    return measure_names



import itertools
import logging
import pprint
import random
import string
import textwrap
from collections import OrderedDict
from collections.abc import Mapping
from dataclasses import is_dataclass, fields
from enum import Enum
from hashlib import sha1
from typing import Sequence, TypeVar, Tuple, NoReturn, Type, Any, List

from metricflow.model.objects.base import HashableBaseModel

logger = logging.getLogger(__name__)


def assert_exactly_one_arg_set(**kwargs) -> None:  # type: ignore
    """Throws an assertion error if 0 or more than 1 argument is not None."""
    num_set = 0
    for value in kwargs.values():
        if value is not None:
            num_set += 1

    assert num_set == 1, f"{num_set} argument(s) set instead of 1 in arguments: {kwargs}"


def is_hashable_base_model(obj):  # type:ignore # noqa: D
    return isinstance(obj, HashableBaseModel)


def _to_pretty_printable_object(obj):  # type: ignore
    """Convert the object that will look nicely when fed into the PrettyPrinter.

    Main change is that dataclasses will have a field with the class name. In Python 3.10, the pretty printer class will
    support dataclasses, so we can remove this once we're on 3.10. Also tried the prettyprint package with dataclasses,
    but that prints full names for the classes e.g. a.b.MyClass and it also always added line breaks, even if an object
    could fit on one line, so preferred to not use that.

    e.g.
    metricflow.specs.DimensionSpec(
        element_name='country',
        identifier_links=()
    ),

    Instead, the below will print something like:

    {'class': 'DimensionSpec',
     'element_name': 'country_latest',
     'identifier_links': ({'class': 'IdentifierSpec',
                           'element_name': 'listing',
                           'identifier_links': ()},)}
    """
    if obj is None:
        return None

    elif isinstance(obj, (str, int, float)):
        return obj

    elif isinstance(obj, (list, tuple)):
        result = []
        for item in obj:
            result.append(_to_pretty_printable_object(item))

        if isinstance(obj, list):
            return result
        elif isinstance(obj, tuple):
            return tuple(result)

        assert False

    elif isinstance(obj, Mapping):
        result = {}
        for key, value in obj.items():
            result[_to_pretty_printable_object(key)] = _to_pretty_printable_object(value)
        return result

    elif is_dataclass(obj):
        result = {"class": type(obj).__name__}

        for field in fields(obj):
            result[field.name] = _to_pretty_printable_object(getattr(obj, field.name))
        return result
    elif is_hashable_base_model(obj):
        result = {"class": type(obj).__name__}

        for field_name, value in obj.dict().items():
            result[field_name] = _to_pretty_printable_object(value)
        return result

    # Can't make it more pretty.
    return obj


def pretty_format(obj) -> str:  # type: ignore
    """Return the object as a string that looks pretty."""
    if isinstance(obj, str):
        return obj
    return pprint.pformat(_to_pretty_printable_object(obj), width=80, sort_dicts=False)


def pformat_big_objects(*args, **kwargs) -> str:  # type: ignore
    """Prints a series of objects with many fields in a pretty way.

    See _to_pretty_printable_object() for more context on this format. Looks like:

    measure_recipe:
    {'class': 'MeasureRecipe',
     'measure_node': ReadSqlSourceNode(node_id=rss_140),
     'required_local_linkable_specs': ({'class': 'DimensionSpec',
                                        'element_name': 'is_instant',
                                        'identifier_links': ()},),
     'join_linkable_instances_recipes': ()}

    """
    items = []
    for arg in args:
        items.append(pretty_format(arg))
    for key, value in kwargs.items():
        items.append(f"{key}:")
        items.append(textwrap.indent(pretty_format(value), prefix="    "))
    return "\n".join(items)


SequenceT = TypeVar("SequenceT")


def flatten_nested_sequence(sequence_of_sequences: Sequence[Sequence[SequenceT]]) -> Tuple[SequenceT, ...]:
    """Convert a nested sequence into a flattened tuple.

    e.g. ((1,2), (3,4)) -> (1, 2, 3, 4)
    """
    return tuple(itertools.chain.from_iterable(sequence_of_sequences))


def flatten_and_dedupe(sequence_of_sequences: Sequence[Sequence[SequenceT]]) -> Tuple[SequenceT, ...]:
    """Convert a nested sequence into a flattened tuple, with de-duping.

    e.g. ((1,2), (2,3)) -> (1, 2, 3)
    """
    items = flatten_nested_sequence(sequence_of_sequences)
    return tuple(OrderedDict.fromkeys(items))


def random_id() -> str:
    """Generates an 8-digit random alphanumeric string."""
    alphabet = string.ascii_lowercase + string.digits
    # Characters that go below the line are visually unappealing, so don't use those.
    filtered_alphabet = [x for x in alphabet if x not in "gjpqy"]
    return "".join(random.choices(filtered_alphabet, k=8))


def assert_values_exhausted(value: NoReturn) -> NoReturn:
    """Helper method to allow MyPy to guarantee an exhaustive switch through an enumeration or literal

    DO NOT MODIFY THE TYPE SIGNATURE OF THIS FUNCTION UNLESS MYPY CHANGES HOW IT HANDLES THINGS

    To use this function correctly you MUST do an exhaustive switch through ALL values, using `is` for comparison
    (doing x == SomeEnum.VALUE will not work, nor will `x in (SomeEnum.VALUE_1, SomeEnum.VALUE_2)`).

    If mypy raises an error of the form:
      `x has incompatible type SomeEnum; expected NoReturn`
    the switch is not constructed correctly. Fix your switch statement to use `is` for all comparisons.

    If mypy raises an error of the form
      `x has incompatible type Union[Literal...]` expected NoReturn`
    the switch statement is non-exhaustive, and the values listed in the error message need to be accounted for.

    See https://mypy.readthedocs.io/en/stable/literal_types.html#exhaustiveness-checks
    For an enum example, see issue:
    https://github.com/python/mypy/issues/6366#issuecomment-560369716
    """
    assert False, f"Should be unreachable, but got {value}"


def hash_strings(strings: Sequence[str]) -> str:
    """Produces a hash from a list of strings."""
    hash_builder = sha1()
    for s in strings:
        hash_builder.update(s.encode("utf-8"))
    return hash_builder.hexdigest()


T = TypeVar("T", bound="ExtendedEnum")


class ExtendedEnum(Enum):
    """Extension of standard Enum class with some extra utilities"""

    @classmethod
    def _missing_(cls: Type[T], value: Any) -> "ExtendedEnum":  # type: ignore[misc]
        """Make enums case insensitive"""
        for member in cls:
            if member.value == value.lower():
                return member
            if member.value == value.upper():
                return member

        raise ValueError(f"Invalid enum value: `{value}` in enum {cls.__name__}")

    @classmethod
    def for_name(cls: Type[T], name: str) -> T:
        """Return enum member with this name"""
        if name not in cls.__members__:
            raise KeyError(f"Unable to find name `{name}` in enum {cls.__name__}")
        return getattr(cls, name)

    @classmethod
    def list_names(cls) -> List[str]:
        """List valid names within this enum class"""
        return list(cls.__members__.keys())



import logging
from dataclasses import dataclass
from typing import Optional, Tuple, Protocol

import pytest

from metricflow.dataclass_serialization import SerializableDataclass, DataClassDeserializer, DataclassSerializer

logger = logging.getLogger(__name__)


@pytest.fixture
def dataclass_serializer() -> DataclassSerializer:  # noqa: D
    return DataclassSerializer()


@pytest.fixture
def dataclass_deserializer() -> DataClassDeserializer:  # noqa: D
    return DataClassDeserializer()


@dataclass(frozen=True)
class SimpleDataclass(SerializableDataclass):  # noqa: D
    field0: int = -1


@dataclass(frozen=True)
class NestedDataclass(SerializableDataclass):  # noqa: D
    field1: SimpleDataclass


@dataclass(frozen=True)
class DeeplyNestedDataclass(SerializableDataclass):  # noqa: D
    field2: NestedDataclass


@dataclass(frozen=True)
class DataclassWithOptional(SerializableDataclass):  # noqa: D
    field3: Optional[SimpleDataclass] = None
    field4: Optional[SimpleDataclass] = None


@dataclass(frozen=True)
class DataclassWithTuple(SerializableDataclass):  # noqa: D
    field5: Tuple[SimpleDataclass, ...]


class SimpleProtocol(Protocol):  # noqa: D
    field6: int


@dataclass(frozen=True)
class SimpleClassWithProtocol(SimpleProtocol, SerializableDataclass):  # noqa: D
    field6: int


@dataclass(frozen=True)
class NestedDataclassWithProtocol(SerializableDataclass):  # noqa: D
    field7: SimpleClassWithProtocol


@dataclass(frozen=True)
class DataclassWithDefaultTuple(SerializableDataclass):  # noqa: D
    field8: Tuple[DataclassWithOptional, ...] = tuple()


@dataclass(frozen=True)
class DataclassWithDataclassDefault(SerializableDataclass):  # noqa: D
    field9: SimpleDataclass = SimpleDataclass(field0=-10)


def test_simple_dataclass(  # noqa: D
    dataclass_serializer: DataclassSerializer, dataclass_deserializer: DataClassDeserializer
) -> None:
    obj = SimpleDataclass(field0=1)
    serialized_object = dataclass_serializer.pydantic_serialize(obj)

    deserialized_object = dataclass_deserializer.pydantic_deserialize(SimpleDataclass, serialized_obj=serialized_object)
    assert obj == deserialized_object


def test_nested_dataclass(  # noqa: D
    dataclass_serializer: DataclassSerializer, dataclass_deserializer: DataClassDeserializer
) -> None:
    obj = NestedDataclass(field1=SimpleDataclass(field0=1))
    serialized_object = dataclass_serializer.pydantic_serialize(obj)
    deserialized_object = dataclass_deserializer.pydantic_deserialize(NestedDataclass, serialized_obj=serialized_object)
    assert obj == deserialized_object


def test_deeply_nested_dataclass(  # noqa: D
    dataclass_serializer: DataclassSerializer, dataclass_deserializer: DataClassDeserializer
) -> None:
    obj = DeeplyNestedDataclass(field2=NestedDataclass(field1=SimpleDataclass(field0=1)))
    serialized_object = dataclass_serializer.pydantic_serialize(obj)

    deserialized_object = dataclass_deserializer.pydantic_deserialize(
        DeeplyNestedDataclass, serialized_obj=serialized_object
    )
    assert obj == deserialized_object


def test_dataclass_with_optional(  # noqa: D
    dataclass_serializer: DataclassSerializer, dataclass_deserializer: DataClassDeserializer
) -> None:
    # DataclassWithOptional has 2 optional fields, so this tests the None and not-None cases.
    obj = DataclassWithOptional(field4=SimpleDataclass(field0=1))
    serialized_object = dataclass_serializer.pydantic_serialize(obj)

    deserialized_object = dataclass_deserializer.pydantic_deserialize(
        DataclassWithOptional, serialized_obj=serialized_object
    )
    assert obj == deserialized_object


def test_dataclass_deserialization_with_optional_default(dataclass_deserializer: DataClassDeserializer) -> None:
    """Tests application of default value when deserializing a sparse optional

    This is necessary to support cases where new optional fields are added and defaulted None, but serialized
    values - which should be compatible - exist without the field in place
    """
    serialized_object = r'{"field4": {"field0": 5}}'

    deserialized_object = dataclass_deserializer.pydantic_deserialize(
        DataclassWithOptional, serialized_obj=serialized_object
    )
    assert deserialized_object == DataclassWithOptional(field4=SimpleDataclass(field0=5))
    # Verify default
    assert deserialized_object.field3 is None


def test_dataclass_with_tuple(  # noqa: D
    dataclass_serializer: DataclassSerializer, dataclass_deserializer: DataClassDeserializer
) -> None:
    obj = DataclassWithTuple(field5=(SimpleDataclass(field0=1),))
    serialized_object = dataclass_serializer.pydantic_serialize(obj)

    deserialized_object = dataclass_deserializer.pydantic_deserialize(
        DataclassWithTuple, serialized_obj=serialized_object
    )
    assert obj == deserialized_object


def test_dataclass_deserialization_with_tuple_default(dataclass_deserializer: DataClassDeserializer) -> None:
    """Tests application of default tuple value when deserializing a sparse tuple dataclass container

    As with Optional, this is necessary to support cases where a default is set for a field added after
    some instances were already serialized, but with an empty tuple in place as a default.
    """
    serialized_object = r"{}"

    deserialized_object = dataclass_deserializer.pydantic_deserialize(
        DataclassWithDefaultTuple, serialized_obj=serialized_object
    )
    assert deserialized_object == DataclassWithDefaultTuple()
    # Verify default
    assert deserialized_object.field8 == tuple()


def test_dataclass_with_protocol(  # noqa: D
    dataclass_serializer: DataclassSerializer, dataclass_deserializer: DataClassDeserializer
) -> None:
    obj = SimpleClassWithProtocol(field6=1)
    serialized_object = dataclass_serializer.pydantic_serialize(obj)

    deserialized_object = dataclass_deserializer.pydantic_deserialize(
        SimpleClassWithProtocol, serialized_obj=serialized_object
    )
    assert obj == deserialized_object


def test_nested_dataclass_with_protocol(  # noqa: D
    dataclass_serializer: DataclassSerializer, dataclass_deserializer: DataClassDeserializer
) -> None:
    obj = NestedDataclassWithProtocol(field7=SimpleClassWithProtocol(field6=1))
    serialized_object = dataclass_serializer.pydantic_serialize(obj)

    deserialized_object = dataclass_deserializer.pydantic_deserialize(
        NestedDataclassWithProtocol, serialized_obj=serialized_object
    )
    assert obj == deserialized_object


def test_nested_dataclass_deserialization_with_primitive_default(dataclass_deserializer: DataClassDeserializer) -> None:
    """Tests application of default value when deserializing a sparse primitive type dataclass container

    This ensures a default set on a primitive, with nesting, still deserializes correctly even if the value is
    not present.
    """
    serialized_object = r'{"field1": {}}'

    deserialized_object = dataclass_deserializer.pydantic_deserialize(NestedDataclass, serialized_obj=serialized_object)
    assert deserialized_object == NestedDataclass(field1=SimpleDataclass())
    # Verify default
    assert deserialized_object.field1.field0 == -1


def test_dataclass_deserialization_with_dataclass_default(dataclass_deserializer: DataClassDeserializer) -> None:
    """Tests application of default value when deserializing a dataclass with a default dataclass value set"""
    serialized_object = r"{}"

    deserialized_object = dataclass_deserializer.pydantic_deserialize(
        DataclassWithDataclassDefault, serialized_obj=serialized_object
    )
    assert deserialized_object == DataclassWithDataclassDefault(field9=SimpleDataclass(field0=-10))
    # Verify default
    assert deserialized_object.field9.field0 == -10



import logging
import textwrap

from metricflow.object_utils import pformat_big_objects
from metricflow.telemetry.handlers.handlers import TelemetryHandler, PayloadType

logger = logging.getLogger(__name__)


class PythonLoggerTelemetryHandler(TelemetryHandler):
    """A telemetry client that logs data to the Python logger for debugging during tests."""

    def __init__(self, logger_level: int) -> None:  # noqa: D
        self._logger_level = logger_level

    def _write_log(self, client_id: str, payload: PayloadType) -> None:  # noqa: D
        logger.log(
            level=self._logger_level,
            msg=f"Logging telemetry payload:\n{textwrap.indent(pformat_big_objects(payload), prefix='    ')}",
        )



import pytest

from metricflow.dataflow.sql_table import SqlTable


def test_sql_table() -> None:  # noqa: D
    sql_table = SqlTable(schema_name="foo", table_name="bar")

    assert sql_table.sql == "foo.bar"
    assert SqlTable.from_string("foo.bar") == sql_table

    json_serialized_table = sql_table.json()
    deserialized_table = SqlTable.parse_raw(json_serialized_table)

    assert sql_table == deserialized_table


def test_sql_table_with_db() -> None:  # noqa: D
    sql_table = SqlTable(db_name="db", schema_name="foo", table_name="bar")

    assert sql_table.sql == "db.foo.bar"
    assert SqlTable.from_string("db.foo.bar") == sql_table

    json_serialized_table = sql_table.json()
    deserialized_table = SqlTable.parse_raw(json_serialized_table)

    assert sql_table == deserialized_table


def test_invalid_sql_table() -> None:  # noqa: D
    with pytest.raises(RuntimeError):
        SqlTable.from_string("foo")



from metricflow.execution.execution_plan import ExecutionPlan
from metricflow.execution.executor import SequentialPlanExecutor
from metricflow.test.execution.noop_task import NoOpExecutionPlanTask


def test_single_task() -> None:
    """Tests running an execution plan with a single task."""
    task = NoOpExecutionPlanTask()
    execution_plan = ExecutionPlan("plan0", leaf_tasks=[task])
    results = SequentialPlanExecutor().execute_plan(execution_plan)
    assert results.get_result(task.task_id)


def test_single_task_error() -> None:
    """Check that an error is properly returned in the results if a task errors out."""
    task = NoOpExecutionPlanTask(should_error=True)
    execution_plan = ExecutionPlan("plan0", leaf_tasks=[task])
    executor = SequentialPlanExecutor()
    results = executor.execute_plan(execution_plan)
    assert len(results.get_result(task.task_id).errors) == 1
    assert results.get_result(task.task_id).errors[0] == NoOpExecutionPlanTask.EXAMPLE_ERROR


def test_task_with_parents() -> None:
    """Tests a plan with a task that has 2 direct parents."""
    parent_task1 = NoOpExecutionPlanTask()
    parent_task2 = NoOpExecutionPlanTask()
    leaf_task = NoOpExecutionPlanTask(parent_tasks=[parent_task1, parent_task2])
    execution_plan = ExecutionPlan("plan0", leaf_tasks=[leaf_task])
    results = SequentialPlanExecutor().execute_plan(execution_plan)

    # Check that they all finished.
    parent_result1 = results.get_result(parent_task1.task_id)
    parent_result2 = results.get_result(parent_task2.task_id)
    leaf_result = results.get_result(leaf_task.task_id)

    # Check that parents completed before the leaf.
    assert parent_result1.end_time < leaf_result.end_time
    assert parent_result2.end_time < leaf_result.end_time

    assert not results.contains_task_errors


def test_parent_task_error() -> None:
    """Check that a child task is not run if a parent task fails."""
    parent_task1 = NoOpExecutionPlanTask(should_error=True)
    parent_task2 = NoOpExecutionPlanTask()
    leaf_task = NoOpExecutionPlanTask(parent_tasks=[parent_task1, parent_task2])
    execution_plan = ExecutionPlan("plan0", leaf_tasks=[leaf_task])

    executor = SequentialPlanExecutor()
    results = executor.execute_plan(execution_plan)
    assert len(results.all_results()) == 1
    assert results.get_result(parent_task1.task_id).errors[0] == NoOpExecutionPlanTask.EXAMPLE_ERROR



from __future__ import annotations

import datetime
import logging
import textwrap
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, List, Sequence

import pandas as pd

from metricflow.configuration.constants import CONFIG_DBT_PROFILE, CONFIG_DBT_REPO, CONFIG_DBT_TARGET, CONFIG_DWH_SCHEMA
from metricflow.configuration.yaml_handler import YamlFileHandler
from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.source_node import SourceNodeBuilder
from metricflow.dataflow.dataflow_plan import DataflowPlan
from metricflow.dataflow.sql_table import SqlTable
from metricflow.dataset.convert_data_source import DataSourceToDataSetConverter
from metricflow.engine.models import Dimension, Materialization, Metric
from metricflow.engine.time_source import ServerTimeSource
from metricflow.engine.utils import build_user_configured_model_from_config
from metricflow.execution.execution_plan import ExecutionPlan, SqlQuery
from metricflow.execution.execution_plan_to_text import execution_plan_to_text
from metricflow.execution.executor import SequentialPlanExecutor
from metricflow.model.semantic_model import SemanticModel
from metricflow.model.semantics.linkable_spec_resolver import LinkableElementProperties
from metricflow.object_utils import pformat_big_objects, random_id
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.dataflow_to_execution import DataflowToExecutionPlanConverter
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter
from metricflow.plan_conversion.time_spine import TimeSpineSource, TimeSpineTableBuilder
from metricflow.protocols.sql_client import SqlClient
from metricflow.query.query_parser import MetricFlowQueryParser
from metricflow.specs import ColumnAssociationResolver, MetricFlowQuerySpec
from metricflow.references import MetricReference
from metricflow.sql.optimizer.optimization_levels import SqlQueryOptimizationLevel
from metricflow.time.time_source import TimeSource
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.errors.errors import ExecutionException, MaterializationNotFoundError
from metricflow.sql_clients.sql_utils import make_sql_client_from_config
from metricflow.sql_clients.common_client import not_empty
from metricflow.telemetry.models import TelemetryLevel
from metricflow.telemetry.reporter import TelemetryReporter, log_call

logger = logging.getLogger(__name__)
_telemetry_reporter = TelemetryReporter(report_levels_higher_or_equal_to=TelemetryLevel.USAGE)
_telemetry_reporter.add_python_log_handler()
_telemetry_reporter.add_rudderstack_handler()


@dataclass(frozen=True)
class MetricFlowRequestId:
    """Uniquely identifies a request to the MF engine."""

    mf_rid: str


@dataclass(frozen=True)
class MetricFlowQueryRequest:
    """Encapsulates the parameters for a metric query.

    metric_names: Names of the metrics to query.
    group_by_names: Names of the dimensions and identifiers to query.
    limit: Limit the result to this many rows.
    time_constraint_start: Get data for the start of this time range.
    time_constraint_end: Get data for the end of this time range.
    where_constraint: A SQL string using group by names that can be used like a where clause on the output data.
    order_by_names: metric and group by names to order by. A "-" can be used to specify reverse order e.g. "-ds"
    output_table: If specified, output the result data to this table instead of a result dataframe.
    sql_optimization_level: The level of optimization for the generated SQL.
    """

    request_id: MetricFlowRequestId
    metric_names: Sequence[str]
    group_by_names: Sequence[str]
    limit: Optional[int] = None
    time_constraint_start: Optional[datetime.datetime] = None
    time_constraint_end: Optional[datetime.datetime] = None
    where_constraint: Optional[str] = None
    order_by_names: Optional[Sequence[str]] = None
    output_table: Optional[str] = None
    sql_optimization_level: SqlQueryOptimizationLevel = SqlQueryOptimizationLevel.O4

    @staticmethod
    def create_with_random_request_id(  # noqa: D
        metric_names: Sequence[str],
        group_by_names: Sequence[str],
        limit: Optional[int] = None,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
        where_constraint: Optional[str] = None,
        order_by_names: Optional[Sequence[str]] = None,
        output_table: Optional[str] = None,
        sql_optimization_level: SqlQueryOptimizationLevel = SqlQueryOptimizationLevel.O4,
    ) -> MetricFlowQueryRequest:
        return MetricFlowQueryRequest(
            request_id=MetricFlowRequestId(mf_rid=f"{random_id()}"),
            metric_names=metric_names,
            group_by_names=group_by_names,
            limit=limit,
            time_constraint_start=time_constraint_start,
            time_constraint_end=time_constraint_end,
            where_constraint=where_constraint,
            order_by_names=order_by_names,
            output_table=output_table,
            sql_optimization_level=sql_optimization_level,
        )


@dataclass(frozen=True)
class MetricFlowQueryResult:  # noqa: D
    """The result of a query and context on how it was generated."""

    query_spec: MetricFlowQuerySpec
    dataflow_plan: DataflowPlan[DataSourceDataSet]
    sql: str
    result_df: Optional[pd.DataFrame] = None
    result_table: Optional[SqlTable] = None


@dataclass(frozen=True)
class MetricFlowExplainResult:
    """Returns plans for resolving a query."""

    query_spec: MetricFlowQuerySpec
    dataflow_plan: DataflowPlan[DataSourceDataSet]
    execution_plan: ExecutionPlan
    output_table: Optional[SqlTable] = None

    @property
    def rendered_sql(self) -> SqlQuery:
        """Return the SQL query that would be run for the given query."""
        if len(self.execution_plan.tasks) != 1:
            raise NotImplementedError(
                f"Multiple tasks in the execution plan not yet supported. Got tasks: {self.execution_plan.tasks}"
            )

        sql_query = self.execution_plan.tasks[0].sql_query
        if not sql_query:
            raise NotImplementedError(
                f"Execution plan tasks without a SQL query not yet supported. Got tasks: {self.execution_plan.tasks}"
            )

        return sql_query

    @property
    def rendered_sql_without_descriptions(self) -> SqlQuery:
        """Return the SQL query without the inline descriptions."""
        sql_query = self.rendered_sql
        return SqlQuery(
            sql_query="\n".join(
                filter(lambda line: not line.strip().startswith("--"), sql_query.sql_query.split("\n"))
            ),
            bind_parameters=sql_query.bind_parameters,
        )


class AbstractMetricFlowEngine(ABC):
    """Query interface for clients"""

    @abstractmethod
    def query(
        self,
        mf_request: MetricFlowQueryRequest,
    ) -> MetricFlowQueryResult:
        """Query for metrics."""
        pass

    @abstractmethod
    def explain(
        self,
        mf_request: MetricFlowQueryRequest,
    ) -> MetricFlowExplainResult:
        """Similar to query - returns the query that would have been executed."""
        pass

    @abstractmethod
    def simple_dimensions_for_metrics(self, metric_names: List[str]) -> List[Dimension]:
        """Retrieves a list of all common dimensions for metric_names.

        "simple" dimensions are the ones that people expect from a UI perspective. For example, if "ds" is a time
        dimension at a day granularity, this would not list "ds__week".

        Args:
            metric_names: Names of metrics to get common dimensions from.

        Returns:
            A list of Dimension objects containing metadata.
        """
        pass

    @abstractmethod
    def list_metrics(self) -> List[Metric]:
        """Retrieves a list of metric names.

        Returns:
            A list of Metric objects containing metadata.
        """
        pass

    @abstractmethod
    def get_dimension_values(
        self,
        metric_name: str,
        get_group_by_values: str,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> List[str]:
        """Retrieves a list of dimension values given a [metric_name, get_group_by_values].

        Args:
            metric_name: Name of metric that contains the group_by.
            get_group_by_values: Name of group_by to get values from.
            time_constraint_start: Get data for the start of this time range.
            time_constraint_end: Get data for the end of this time range.

        Returns:
            A list of dimension values as string.
        """
        pass

    @abstractmethod
    def list_materializations(self) -> List[Materialization]:
        """Retrieves a list of materialization names.

        Returns:
            A list of Materialization objects containing metadata.
        """
        pass

    @abstractmethod
    def materialize(
        self,
        materialization_name: str,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> SqlTable:
        """Builds a table containing metrics and dimensions from a materialization definition.

        This can be very expensive if a large time range is provided.

        Args:
            materialization_name: Name of materialization
            time_constraint_start: Materialized for the start of this time range.
            time_constraint_end: Materialized for the end of this time range.

        Returns:
            SqlTable object of the materialized table.
        """
        pass

    @abstractmethod
    def drop_materialization(self, materialization_name: str) -> bool:
        """Drops the table associated with a materialization definition.

        Args:
            materialization_name: Name of materialization to drop.

        Returns:
            True if a table has been drop, False if table doesn't exist.
        """
        pass


class MetricFlowEngine(AbstractMetricFlowEngine):
    """Main entry point for queries."""

    _LOGGING_INDENT = "   "

    @staticmethod
    def from_config(handler: YamlFileHandler) -> MetricFlowEngine:
        """Initialize MetricFlowEngine via yaml config file."""
        sql_client = make_sql_client_from_config(handler)

        # Ideally we should put this getting of of CONFIG_DBT_REPO in a helper
        dbt_repo = handler.get_value(CONFIG_DBT_REPO) or ""
        if dbt_repo.lower() in ["yes", "y", "true", "t", "1"]:
            # This import results in eventually importing dbt, and dbt is an
            # optional dep meaning it isn't guaranteed to be installed. If the
            # import is at the top ofthe file MetricFlow will blow up if dbt
            # isn't installed. Thus by importing it here, we only run into the
            # exception if this conditional is hit without dbt installed
            from metricflow.engine.utils import build_user_configured_model_from_dbt_config

            dbt_profile = handler.get_value(CONFIG_DBT_PROFILE)
            dbt_target = handler.get_value(CONFIG_DBT_TARGET)

            semantic_model = SemanticModel(
                build_user_configured_model_from_dbt_config(handler=handler, profile=dbt_profile, target=dbt_target)
            )
        else:
            semantic_model = SemanticModel(build_user_configured_model_from_config(handler))
        system_schema = not_empty(handler.get_value(CONFIG_DWH_SCHEMA), CONFIG_DWH_SCHEMA, handler.url)
        return MetricFlowEngine(
            semantic_model=semantic_model,
            sql_client=sql_client,
            system_schema=system_schema,
        )

    def __init__(
        self,
        semantic_model: SemanticModel,
        sql_client: SqlClient,
        system_schema: str,
        time_source: TimeSource = ServerTimeSource(),
        column_association_resolver: Optional[ColumnAssociationResolver] = None,
        time_spine_source: Optional[TimeSpineSource] = None,
    ) -> None:
        """Initializer for MetricFlowEngine

        For direct calls to construct MetricFlowEngine, do not pass the following parameters,
        - time_source
        - column_association_resolver
        - time_spine_source
        These parameters are mainly there to be overridden during tests.
        """

        self._semantic_model = semantic_model
        self._sql_client = sql_client
        self._column_association_resolver = column_association_resolver or (
            DefaultColumnAssociationResolver(semantic_model)
        )
        self._time_source = time_source
        self._time_spine_source = time_spine_source or TimeSpineSource(schema_name=system_schema)
        self._time_spine_table_builder = TimeSpineTableBuilder(
            time_spine_source=self._time_spine_source, sql_client=self._sql_client
        )

        self._schema = system_schema

        self._source_data_sets: List[DataSourceDataSet] = []
        converter = DataSourceToDataSetConverter(column_association_resolver=self._column_association_resolver)
        for data_source in self._semantic_model.user_configured_model.data_sources:
            data_set = converter.create_sql_source_data_set(data_source)
            self._source_data_sets.append(data_set)
            logger.info(f"Created source dataset from data source '{data_source.name}'")

        source_node_builder = SourceNodeBuilder(self._semantic_model)
        source_nodes = source_node_builder.create_from_data_sets(self._source_data_sets)

        node_output_resolver = DataflowPlanNodeOutputDataSetResolver[DataSourceDataSet](
            column_association_resolver=DefaultColumnAssociationResolver(semantic_model),
            semantic_model=semantic_model,
            time_spine_source=self._time_spine_source,
        )

        self._dataflow_plan_builder = DataflowPlanBuilder[DataSourceDataSet](
            source_nodes=source_nodes,
            semantic_model=self._semantic_model,
            time_spine_source=self._time_spine_source,
        )
        self._to_sql_query_plan_converter = DataflowToSqlQueryPlanConverter[DataSourceDataSet](
            column_association_resolver=self._column_association_resolver,
            semantic_model=self._semantic_model,
            time_spine_source=self._time_spine_source,
        )
        self._to_execution_plan_converter = DataflowToExecutionPlanConverter[DataSourceDataSet](
            sql_plan_converter=self._to_sql_query_plan_converter,
            sql_plan_renderer=self._sql_client.sql_engine_attributes.sql_query_plan_renderer,
            sql_client=sql_client,
        )
        self._executor = SequentialPlanExecutor()

        self._query_parser = MetricFlowQueryParser(
            model=self._semantic_model,
            source_nodes=source_nodes,
            node_output_resolver=node_output_resolver,
        )

    def _get_materialization_by_name(self, materialization_name: str) -> Optional[Materialization]:
        materializations = self.list_materializations()
        for mat in materializations:
            if mat.name == materialization_name:
                return mat
        return None

    def _generate_sql_table(self, table_name: str) -> SqlTable:
        return SqlTable.from_string(f"{self._schema}.{table_name}")

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def query(self, mf_request: MetricFlowQueryRequest) -> MetricFlowQueryResult:  # noqa: D
        logger.info(
            f"Starting query request:\n"
            f"{textwrap.indent(pformat_big_objects(mf_request), prefix=MetricFlowEngine._LOGGING_INDENT)}"
        )
        explain_result = self._create_execution_plan(mf_request)
        execution_plan = explain_result.execution_plan

        if len(execution_plan.tasks) != 1:
            raise NotImplementedError("Multiple tasks not yet supported.")

        task = execution_plan.tasks[0]

        logger.info(f"Sequentially running tasks in:\n" f"{execution_plan_to_text(execution_plan)}")
        execution_results = self._executor.execute_plan(execution_plan)
        logger.info("Finished running tasks in execution plan")

        if execution_results.contains_task_errors:
            raise ExecutionException(f"Got errors while executing tasks:\n{execution_results.get_result(task.task_id)}")

        task_execution_result = execution_results.get_result(task.task_id)

        assert task_execution_result.sql, "Task execution should have returned SQL that was run"

        logger.info(f"Finished query request: {mf_request.request_id}")
        return MetricFlowQueryResult(
            query_spec=explain_result.query_spec,
            dataflow_plan=explain_result.dataflow_plan,
            sql=task_execution_result.sql,
            result_df=task_execution_result.df,
            result_table=explain_result.output_table,
        )

    def _create_execution_plan(self, mf_query_request: MetricFlowQueryRequest) -> MetricFlowExplainResult:
        query_spec = self._query_parser.parse_and_validate_query(
            metric_names=mf_query_request.metric_names,
            group_by_names=mf_query_request.group_by_names,
            limit=mf_query_request.limit,
            time_constraint_start=mf_query_request.time_constraint_start,
            time_constraint_end=mf_query_request.time_constraint_end,
            where_constraint_str=mf_query_request.where_constraint,
            order=mf_query_request.order_by_names,
        )
        logger.info(f"Query spec is:\n{pformat_big_objects(query_spec)}")

        if self._semantic_model.metric_semantics.contains_cumulative_metric(
            tuple(m.as_reference for m in query_spec.metric_specs)
        ):
            self._time_spine_table_builder.create_if_necessary()
            time_constraint_updated = False
            if not mf_query_request.time_constraint_start:
                time_constraint_start = self._time_source.get_time() - datetime.timedelta(days=365)
                logger.warning(
                    "A start time has not be supplied while querying for cumulative metrics. To avoid an excessive "
                    f"number of rows, the start time will be changed to {time_constraint_start.isoformat()}"
                )
                time_constraint_updated = True
            if not mf_query_request.time_constraint_end:
                time_constraint_end = self._time_source.get_time()
                logger.warning(
                    "A end time has not be supplied while querying for cumulative metrics. To avoid an excessive "
                    f"number of rows, the end time will be changed to {time_constraint_end.isoformat()}"
                )
                time_constraint_updated = True
            if time_constraint_updated:
                query_spec = self._query_parser.parse_and_validate_query(
                    metric_names=mf_query_request.metric_names,
                    group_by_names=mf_query_request.group_by_names,
                    limit=mf_query_request.limit,
                    time_constraint_start=mf_query_request.time_constraint_start,
                    time_constraint_end=mf_query_request.time_constraint_end,
                    where_constraint_str=mf_query_request.where_constraint,
                    order=mf_query_request.order_by_names,
                )
                logger.warning(f"Query spec updated to:\n{pformat_big_objects(query_spec)}")

        output_table: Optional[SqlTable] = None
        if mf_query_request.output_table is not None:
            output_table = SqlTable.from_string(mf_query_request.output_table)

        dataflow_plan = self._dataflow_plan_builder.build_plan(query_spec, output_table)

        if len(dataflow_plan.sink_output_nodes) > 1:
            raise NotImplementedError(
                f"Multiple output nodes in the dataflow plan not yet supported. "
                f"Got tasks: {dataflow_plan.sink_output_nodes}"
            )

        execution_plan = self._to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)

        return MetricFlowExplainResult(
            query_spec=query_spec,
            dataflow_plan=dataflow_plan,
            execution_plan=execution_plan,
            output_table=output_table,
        )

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def explain(self, mf_request: MetricFlowQueryRequest) -> MetricFlowExplainResult:  # noqa: D
        return self._create_execution_plan(mf_request)

    def simple_dimensions_for_metrics(self, metric_names: List[str]) -> List[Dimension]:  # noqa: D
        return [
            Dimension(name=dim.qualified_name)
            for dim in self._semantic_model.metric_semantics.element_specs_for_metrics(
                metric_references=[MetricReference(element_name=mname) for mname in metric_names],
                without_any_property=frozenset(
                    {
                        LinkableElementProperties.IDENTIFIER,
                        LinkableElementProperties.DERIVED_TIME_GRANULARITY,
                        LinkableElementProperties.LOCAL_LINKED,
                    }
                ),
            )
        ]

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def list_materializations(self) -> List[Materialization]:  # noqa: D
        return [
            Materialization(
                name=mat.name, metrics=mat.metrics, dimensions=mat.dimensions, destination_table=mat.destination_table
            )
            for mat in self._semantic_model.user_configured_model.materializations
        ]

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def list_metrics(self) -> List[Metric]:  # noqa: D
        metric_references = self._semantic_model.metric_semantics.metric_references
        metrics = self._semantic_model.metric_semantics.get_metrics(metric_references)
        return [
            Metric(
                name=metric.name,
                dimensions=self.simple_dimensions_for_metrics([metric.name]),
            )
            for metric in metrics
        ]

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def get_dimension_values(  # noqa: D
        self,
        metric_name: str,
        get_group_by_values: str,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> List[str]:

        # Run query
        query_result = self.query(
            MetricFlowQueryRequest.create_with_random_request_id(
                metric_names=[metric_name],
                group_by_names=[get_group_by_values],
                time_constraint_start=time_constraint_start,
                time_constraint_end=time_constraint_end,
            )
        )
        result_dataframe = query_result.result_df
        if result_dataframe is None:
            return []

        # Process the dimension values
        result_dataframe.dropna(inplace=True)
        dimension_values = [str(val) for val in result_dataframe[get_group_by_values]]
        return dimension_values

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def materialize(  # noqa: D
        self,
        materialization_name: str,
        time_constraint_start: Optional[datetime.datetime] = None,
        time_constraint_end: Optional[datetime.datetime] = None,
    ) -> SqlTable:
        materialization = self._get_materialization_by_name(materialization_name)
        if materialization is None:
            raise MaterializationNotFoundError(
                f"Unable to find materialization `{materialization_name}`. Perhaps it has not been registered"
            )

        # Use destination_table if exists else materialization_name
        output_table = materialization.destination_table or self._generate_sql_table(materialization_name)
        self._sql_client.drop_table(output_table)

        # Executes the query with output_table
        query_result = self.query(
            MetricFlowQueryRequest.create_with_random_request_id(
                metric_names=materialization.metrics,
                group_by_names=materialization.dimensions,
                time_constraint_start=time_constraint_start,
                time_constraint_end=time_constraint_end,
                output_table=output_table.sql,
            )
        )
        assert query_result.result_table
        return query_result.result_table

    @log_call(module_name=__name__, telemetry_reporter=_telemetry_reporter)
    def drop_materialization(self, materialization_name: str) -> bool:  # noqa: D
        materialization = self._get_materialization_by_name(materialization_name)
        if materialization is None:
            raise MaterializationNotFoundError(
                f"Unable to find materialization `{materialization_name}`. Perhaps it has not been registered"
            )

        table = materialization.destination_table or self._generate_sql_table(materialization_name)

        if self._sql_client.table_exists(table):
            self._sql_client.drop_table(table)
            return True
        return False



from enum import Enum
from typing import Sequence

from metricflow.sql.optimizer.column_pruner import SqlColumnPrunerOptimizer
from metricflow.sql.optimizer.rewriting_sub_query_reducer import SqlRewritingSubQueryReducer
from metricflow.sql.optimizer.sql_query_plan_optimizer import SqlQueryPlanOptimizer
from metricflow.sql.optimizer.sub_query_reducer import SqlSubQueryReducer
from metricflow.sql.optimizer.table_alias_simplifier import SqlTableAliasSimplifier


class SqlQueryOptimizationLevel(Enum):
    """Defines the level of query optimization and the associated optimizers to apply."""

    O0 = "O0"
    O1 = "O1"
    O2 = "O2"
    O3 = "O3"
    O4 = "O4"


class SqlQueryOptimizerConfiguration:
    """Defines the different optimizers that should be used at each level"""

    @staticmethod
    def optimizers_for_level(
        level: SqlQueryOptimizationLevel, use_column_alias_in_group_by: bool
    ) -> Sequence[SqlQueryPlanOptimizer]:
        """Return the optimizers that should be applied (in order) for each level"""
        if level is SqlQueryOptimizationLevel.O0:
            return ()
        elif level is SqlQueryOptimizationLevel.O1:
            return (SqlTableAliasSimplifier(),)
        elif level is SqlQueryOptimizationLevel.O2:
            return (SqlColumnPrunerOptimizer(), SqlTableAliasSimplifier())
        elif level is SqlQueryOptimizationLevel.O3:
            return (SqlColumnPrunerOptimizer(), SqlSubQueryReducer(), SqlTableAliasSimplifier())
        elif level is SqlQueryOptimizationLevel.O4:
            return (
                SqlColumnPrunerOptimizer(),
                SqlRewritingSubQueryReducer(use_column_alias_in_group_bys=use_column_alias_in_group_by),
                SqlTableAliasSimplifier(),
            )



from metricflow.dataflow.sql_column import SqlColumn
from metricflow.dataflow.sql_table import SqlTable
from metricflow.inference.context.data_warehouse import (
    ColumnProperties,
    DataWarehouseInferenceContext,
    InferenceColumnType,
    TableProperties,
)
from metricflow.inference.models import InferenceSignalType
import metricflow.inference.rule.defaults as defaults


def get_column_properties(column_str: str, type: InferenceColumnType, unique: bool) -> ColumnProperties:  # noqa: D
    return ColumnProperties(
        column=SqlColumn.from_string(column_str),
        type=type,
        row_count=10000,
        distinct_row_count=10000 if unique else 9000,
        is_nullable=False,
        null_count=0,
        min_value=0,
        max_value=9999,
    )


def test_any_identifier_by_name_matcher() -> None:  # noqa: D
    assert defaults.AnyIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.id", InferenceColumnType.INTEGER, True)
    )
    assert defaults.AnyIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.tableid", InferenceColumnType.INTEGER, True)
    )
    assert defaults.AnyIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.table_id", InferenceColumnType.INTEGER, True)
    )
    assert defaults.AnyIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.othertable_id", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.AnyIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.whatever", InferenceColumnType.INTEGER, True)
    )


def test_primary_identifier_by_name_matcher() -> None:  # noqa: D
    assert defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.id", InferenceColumnType.INTEGER, True)
    )
    assert defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.tableid", InferenceColumnType.INTEGER, True)
    )
    assert defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.table_id", InferenceColumnType.INTEGER, True)
    )
    assert defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.tables.table_id", InferenceColumnType.INTEGER, True)
    )
    assert defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.tables.tableid", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.othertable_id", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.othertableid", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.PrimaryIdentifierByNameRule().match_column(
        get_column_properties("db.schema.table.whatever", InferenceColumnType.INTEGER, True)
    )


def test_unique_identifier_by_distinct_count_matcher() -> None:  # noqa: D
    assert defaults.UniqueIdentifierByDistinctCountRule().match_column(
        get_column_properties("db.schema.table.unique_id", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.UniqueIdentifierByDistinctCountRule().match_column(
        get_column_properties("db.schema.table.unique_id", InferenceColumnType.STRING, False)
    )


def test_time_dimension_by_time_type_matcher() -> None:  # noqa: D
    assert defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.DATETIME, True)
    )

    assert not defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.FLOAT, True)
    )
    assert not defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.BOOLEAN, True)
    )
    assert not defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.STRING, True)
    )
    assert not defaults.TimeDimensionByTimeTypeRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.UNKNOWN, True)
    )


def test_primary_time_dimension_by_name_matcher() -> None:  # noqa: D
    assert defaults.PrimaryTimeDimensionByNameRule().match_column(
        get_column_properties("db.schema.table.ds", InferenceColumnType.DATETIME, True)
    )
    assert defaults.PrimaryTimeDimensionByNameRule().match_column(
        get_column_properties("db.schema.table.created_at", InferenceColumnType.DATETIME, True)
    )
    assert not defaults.PrimaryTimeDimensionByNameRule().match_column(
        get_column_properties("db.schema.table.bla", InferenceColumnType.DATETIME, True)
    )
    assert not defaults.PrimaryTimeDimensionByNameRule().match_column(
        get_column_properties("db.schema.table.time", InferenceColumnType.DATETIME, True)
    )


def test_primary_time_dimension_if_only_time_rule() -> None:  # noqa: D
    table = SqlTable.from_string("db.schema.table")
    single_time_col_warehouse = DataWarehouseInferenceContext(
        table_props=[
            TableProperties(
                table=table,
                column_props=[
                    get_column_properties("db.schema.table.id", InferenceColumnType.INTEGER, True),
                    get_column_properties("db.schema.table.time", InferenceColumnType.DATETIME, True),
                ],
            )
        ]
    )
    single_time_col_signals = defaults.PrimaryTimeDimensionIfOnlyTimeRule().process(single_time_col_warehouse)
    assert len(single_time_col_signals) == 1
    assert single_time_col_signals[0].column == SqlColumn.from_string("db.schema.table.time")
    assert single_time_col_signals[0].type_node == InferenceSignalType.DIMENSION.PRIMARY_TIME

    many_time_col_warehouse = DataWarehouseInferenceContext(
        table_props=[
            TableProperties(
                table=table,
                column_props=[
                    get_column_properties("db.schema.table.id", InferenceColumnType.INTEGER, True),
                    get_column_properties("db.schema.table.time", InferenceColumnType.DATETIME, True),
                    get_column_properties("db.schema.table.othertime", InferenceColumnType.DATETIME, True),
                ],
            )
        ]
    )
    many_time_col_signals = defaults.PrimaryTimeDimensionIfOnlyTimeRule().process(many_time_col_warehouse)
    assert len(many_time_col_signals) == 0


def test_categorical_dimension_by_boolean_type_matcher() -> None:  # noqa: D
    assert defaults.CategoricalDimensionByBooleanTypeRule().match_column(
        get_column_properties("db.schema.table.dim", InferenceColumnType.BOOLEAN, True)
    )
    assert not defaults.CategoricalDimensionByBooleanTypeRule().match_column(
        get_column_properties("db.schema.table.bla", InferenceColumnType.FLOAT, True)
    )


def test_categorical_dimension_by_string_type_matcher() -> None:  # noqa: D
    assert defaults.CategoricalDimensionByStringTypeRule().match_column(
        get_column_properties("db.schema.table.dim", InferenceColumnType.STRING, True)
    )
    assert not defaults.CategoricalDimensionByStringTypeRule().match_column(
        get_column_properties("db.schema.table.bla", InferenceColumnType.FLOAT, True)
    )


def test_categorical_dimension_by_string__and_cardinality_type_matcher() -> None:  # noqa: D
    """Tests the composite of string type and cardinality below supplied threshold

    Since the helper cardinality ratio is always either 1 or 0.9, the cardinality thresholds are set to either above
    0.9 (for checks which should match) or below 0.9 (for checks which should not match, or where the match does
    not matter)
    """

    assert defaults.CategoricalDimensionByStringTypeAndLowCardinalityRule(0.99).match_column(
        get_column_properties("db.schema.table.low_cardinality_string_col", InferenceColumnType.STRING, unique=False)
    )
    # INTEGER type columns never match this rule
    assert not defaults.CategoricalDimensionByStringTypeAndLowCardinalityRule(0.99).match_column(
        get_column_properties("db.schema.table.int_col", InferenceColumnType.INTEGER, unique=False)
    )
    assert not defaults.CategoricalDimensionByCardinalityRatioRule(0.40).match_column(
        get_column_properties("db.schema.table.high_cardinality_string_col", InferenceColumnType.STRING, unique=False)
    )


def test_categorical_dimension_by_integer_type_matcher() -> None:  # noqa: D
    assert defaults.CategoricalDimensionByIntegerTypeRule().match_column(
        get_column_properties("db.schema.table.dim", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.CategoricalDimensionByIntegerTypeRule().match_column(
        get_column_properties("db.schema.table.bla", InferenceColumnType.FLOAT, True)
    )


def test_measure_by_real_type_matcher() -> None:  # noqa: D
    assert defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.FLOAT, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.DATETIME, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.BOOLEAN, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.STRING, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.UNKNOWN, True)
    )


def test_measure_by_integer_type_matcher() -> None:  # noqa: D
    assert defaults.MeasureByIntegerTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.INTEGER, True)
    )
    assert not defaults.MeasureByRealTypeRule().match_column(
        get_column_properties("db.schema.table.measure", InferenceColumnType.BOOLEAN, True)
    )



import datetime

import pandas as pd
import pytest

from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql_clients.sql_utils import make_df
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.time.time_constants import ISO8601_PYTHON_FORMAT
from metricflow.time.time_granularity import TimeGranularity

DEFAULT_DS = "ds"


def create_table(sql_client: SqlClient, sql_table: SqlTable, df: pd.DataFrame) -> None:  # noqa: D
    # TODO: Replace with table_exists() check.
    sql_client.drop_table(sql_table)

    sql_client.create_table_from_dataframe(
        sql_table=sql_table,
        df=df,
    )


@pytest.fixture(scope="session")
def create_simple_model_tables(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> bool:
    """Creates tables with example source data for testing."""
    schema = mf_test_session_state.mf_source_schema

    cote_divoire = "cote d'ivoire"

    # NOTE: fct_bookings should have both a host and guest id, but we do not yet
    # have "entity roles" so that a user dimension can be attached to either
    # guest or host (e.g. bookings by user:guest/country)
    data = [
        # While datetime columns can be used here, somewhere there is a conversion to a string when testing
        # with sqlite, so when you read the datetime column, it comes out as a string.
        # Columns for reference
        # ["guest_id", "host_id", "listing_id", "booking_value", "is_instant", "ds", "ds_partitioned",
        #  "booking_paid_at", "referrer_id"]
        ("u0004114", "u0003141", "l3141592", 742.42, False, "2020-01-01", "2020-01-01", "2020-01-02", None),
        ("u0004114", "u0003141", "l3141592", 441.14, True, "2020-01-01", "2020-01-01", "2020-01-02", "u0003141"),
        ("u0003141", "u0003154", "l2718281", 499.99, True, "2020-01-01", "2020-01-01", "2020-01-02", None),
        ("u0003154", "u0003141", "l2718281", 719.89, False, "2020-01-01", "2020-01-01", "2020-01-02", None),
        ("u0004114", "u0003141", "l3141592", 319.85, True, "2020-01-01", "2020-01-01", "2020-01-02", "u0005432"),
        ("u0004114", "u0003141", "l2718281", 442.42, True, "2020-01-02", "2020-01-02", "2020-01-03", None),
        ("u0004114", "u0003141", "l2718281", 241.14, True, "2020-01-02", "2020-01-02", "2020-01-03", None),
        ("u0004114", "u0003141", "l3141592", 799.99, False, "2020-01-02", "2020-01-02", "2020-01-03", None),
        ("u0005432", "u0003452", "l2718281", 319.85, True, "2020-01-02", "2020-01-02", "2020-01-03", "u0003452"),
        ("u0003452", "u0005432", "l3141592", 519.89, False, "2020-01-02", "2020-01-02", "2020-01-03", "u0003141"),
        ("u0003452", "u0004114", "l5948301", 332.23, False, "2020-01-02", "2020-01-02", "2020-01-03", "u0003141"),
        ("u0003452", "u0004114", "l5948301", 0.0, False, "2020-01-03", "2020-01-03", "2020-01-04", None),
    ]
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="fct_bookings"),
        df=make_df(
            sql_client=sql_client,
            columns=[
                "guest_id",
                "host_id",
                "listing_id",
                "booking_value",
                "is_instant",
                DEFAULT_DS,
                "ds_partitioned",
                "booking_paid_at",
                "referrer_id",
            ],
            time_columns={DEFAULT_DS, "ds_partitioned", "booking_paid_at"},
            data=data,
        ),
    )

    # Like fct_bookings, but using dt instead of ds to verify that a primary time dimension named something other than
    # ds can be used.
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="fct_bookings_dt"),
        df=make_df(
            sql_client=sql_client,
            columns=[
                "guest_id",
                "host_id",
                "listing_id",
                "booking_value",
                "is_instant",
                "dt",
                "dt_partitioned",
                "booking_paid_at",
                "referrer_id",
            ],
            time_columns={"dt", "dt_partitioned"},
            data=data,
        ),
    )

    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="fct_views"),
        df=make_df(
            sql_client=sql_client,
            columns=["user_id", "listing_id", DEFAULT_DS, "ds_partitioned"],
            time_columns={DEFAULT_DS, "ds_partitioned"},
            data=[
                ("u0004114", "l3141592", "2020-01-01", "2020-01-01"),
                ("u1612112", "l3141592", "2020-01-01", "2020-01-01"),
                ("u0004114", "l2718281", "2020-01-02", "2020-01-02"),
                ("u0004114", "l3141592", "2020-01-02", "2020-01-02"),
                ("u1612112", "l2718281", "2020-01-02", "2020-01-02"),
            ],
        ),
    )

    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="dim_listings_latest"),
        df=make_df(
            sql_client=sql_client,
            columns=["listing_id", "country", "capacity", "is_lux", "user_id", "created_at"],
            time_columns={"created_at"},
            data=[
                ("l3141592", "us", 3, True, "u0004114", "2020-01-01"),
                ("l5948301", "us", 5, True, "u0004114", "2020-01-02"),
                ("l2718281", cote_divoire, 4, False, "u0005432", "2020-01-02"),
            ],
        ),
    )

    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="dim_listings"),
        df=make_df(
            sql_client=sql_client,
            columns=["listing_id", "country", "capacity", "is_lux", DEFAULT_DS],
            time_columns={DEFAULT_DS},
            data=[
                ("l3141592", "us", 2, True, "2020-01-01"),
                ("l2718281", cote_divoire, 4, True, "2020-01-01"),
                ("l3141592", "us", 3, True, "2020-01-02"),
                ("l2718281", cote_divoire, 4, False, "2020-01-02"),
                ("l5948301", "us", 5, True, "2020-01-02"),
            ],
        ),
    )
    create_table(
        sql_client=sql_client,  # successful identity verifications
        sql_table=SqlTable(schema_name=schema, table_name="fct_id_verifications"),
        df=make_df(
            sql_client=sql_client,
            columns=["verification_id", "user_id", "verification_type", DEFAULT_DS, "ds_partitioned"],
            time_columns={DEFAULT_DS, "ds_partitioned"},
            data=[
                ("v0000001", "u0004114", "drivers_license", "2020-01-01", "2020-01-01"),
                ("v0000002", "u1612112", "passport", "2020-01-01", "2020-01-01"),
                ("v0000003", "u0004114", "social_security", "2020-01-02", "2020-01-02"),
                ("v0000004", "u1612112", "drivers_license", "2020-01-02", "2020-01-02"),
            ],
        ),
    )

    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="dim_users"),
        df=make_df(
            sql_client=sql_client,
            columns=["user_id", "home_state", "created_at", DEFAULT_DS, "ds_partitioned"],
            time_columns={"created_at", DEFAULT_DS, "ds_partitioned"},
            data=[
                ("u0004114", "CA", "2019-03-03", "2020-01-01", "2020-01-01"),
                ("u1612112", "CA", "2019-04-03", "2020-01-01", "2020-01-01"),
                ("u0005432", "TX", "2017-03-03", "2020-01-01", "2020-01-01"),
                ("u0003452", "HI", "2013-03-03", "2020-01-01", "2020-01-01"),
                ("u0003154", "MD", "2014-03-03", "2020-01-01", "2020-01-01"),
                ("u0003141", "NY", "2015-03-03", "2020-01-01", "2020-01-01"),
                ("u0004114", "CA", "2019-03-03", "2020-01-02", "2020-01-02"),
                ("u1612112", "WA", "2019-04-03", "2020-01-02", "2020-01-02"),
                ("u0005432", "TX", "2017-03-03", "2020-01-02", "2020-01-02"),
                ("u0003452", "HI", "2013-03-03", "2020-01-02", "2020-01-02"),
                ("u0003154", "MD", "2014-03-03", "2020-01-02", "2020-01-02"),
                ("u0003141", "NY", "2015-03-03", "2020-01-02", "2020-01-02"),
            ],
        ),
    )

    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="dim_users_latest"),
        df=make_df(
            sql_client=sql_client,
            columns=["user_id", "home_state_latest", DEFAULT_DS],
            time_columns={DEFAULT_DS},
            data=[
                ("u0004114", "CA", "2020-01-02"),
                ("u1612112", "WA", "2020-01-02"),
                ("u0005432", "TX", "2020-01-02"),
                ("u0003452", "HI", "2020-01-02"),
                ("u0003154", "MD", "2020-01-02"),
                ("u0003141", "NY", "2020-01-02"),
            ],
        ),
    )

    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="fct_revenue"),
        df=make_df(
            sql_client=sql_client,
            columns=["revenue", "created_at", "user_id"],
            time_columns={"created_at"},
            data=[
                (1000, "2020-01-01", "u0004114"),
                (4000, "2020-02-01", "u0004114"),
                (3000, "2020-03-01", "u0005432"),
                (4000, "2020-04-02", "u0004114"),
                (1000, "2021-01-01", "u0004114"),
                (1000, "2021-01-02", "u0004114"),
                (1000, "2021-01-03", "u0004114"),
                (1000, "2021-01-04", "u0004114"),
            ],
        ),
    )

    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="dim_lux_listing_id_mapping"),
        df=make_df(
            sql_client=sql_client,
            columns=["listing_id", "lux_listing_id"],
            data=[
                ("l3141592", "ll_001"),
                ("l5948301", "LUX_TEST_ID"),
            ],
        ),
    )

    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="dim_companies"),
        df=make_df(
            sql_client=sql_client,
            columns=["company_id", "user_id", "company_name"],
            data=[
                ("cpid_0", "u0003154", "MD Vacation Rentals LLC"),
            ],
        ),
    )

    accounts_data = [
        ("u0004114", 2135, "2020-01-01", "checking"),
        ("u0004114", 1234, "2020-01-01", "savings"),
        ("u0005432", 5234, "2020-01-03", "checking"),
        ("u0005432", 234, "2020-01-03", "savings"),
        ("u0003141", 24634, "2020-01-01", "checking"),
        ("u0003154", 23452, "2020-01-01", "checking"),
        ("u0004114", 1213, "2020-01-06", "savings"),
        ("u1612112", 5123, "2020-01-02", "checking"),
        ("u0004114", 523, "2020-01-07", "checking"),
        ("u0004114", 7434, "2020-01-10", "checkings"),
        ("u0005432", 8456, "2020-01-02", "savings"),
        ("u0003141", 12939, "2020-01-12", "checking"),
        ("u0003452", 6939, "2020-01-02", "checking"),
        ("u0003452", 35, "2020-01-12", "checking"),
        ("u0005414", 5582, "2020-01-04", "savings"),
    ]

    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="fct_accounts"),
        df=make_df(
            sql_client=sql_client,
            columns=["user_id", "account_balance", DEFAULT_DS, "account_type"],
            time_columns={DEFAULT_DS},
            data=accounts_data,
        ),
    )

    return True


@pytest.fixture(scope="session")
def create_bridge_table(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> bool:
    """Creates tables with example source data for multi-hop join testing."""
    schema = mf_test_session_state.mf_source_schema

    # NOTE: fct_bookings should have both a host and guest id, but we do not yet
    # have "entity roles" so that a user dimension can be attached to either
    # guest or host (e.g. bookings by user:guest/country)
    customer_accounts = [
        # ["customer_id", "customer_name", "customer_atomic_weight", "ds_partitioned"]
        ("0", "thorium", "90", "2020-01-01"),
        ("1", "osmium", "76", "2020-01-02"),
        ("2", "tellurium", "52", "2020-01-01"),
        ("3", "einsteinium", "99", "2020-01-04"),
    ]
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="customer_table"),
        df=make_df(
            sql_client=sql_client,
            columns=["customer_id", "customer_name", "customer_atomic_weight", "ds_partitioned"],
            time_columns={"ds_partitioned"},
            data=customer_accounts,
        ),
    )

    customer_other_data = [
        # ["customer_id", "country", "customer_third_hop_id"]
        ("0", "turkmenistan", "another_id0"),
        ("1", "paraguay", "another_id1"),
        ("2", "myanmar", "another_id2"),
        ("3", "djibouti", "another_id3"),
    ]
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="customer_other_data"),
        df=make_df(
            sql_client=sql_client,
            columns=["customer_id", "country", "customer_third_hop_id"],
            data=customer_other_data,
        ),
    )

    third_hop_table = [
        # ["customer_third_hop_id", "value"]
        ("another_id0", "citadel"),
        ("another_id1", "virtu"),
        ("another_id2", "two sigma"),
        ("another_id3", "jump"),
    ]
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="third_hop_table"),
        df=make_df(
            sql_client=sql_client,
            columns=["customer_third_hop_id", "value"],
            data=third_hop_table,
        ),
    )

    bridge_data = [
        # ["account_id", "customer_id", "extra_dim", "ds_partitioned"]
        ("a0", "0", "lux", "2020-01-01"),
        ("a0", "1", "lux", "2020-01-02"),
        ("a1", "2", "not_lux", "2020-01-01"),
        ("a2", "3", "super_lux", "2020-01-04"),
    ]
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="bridge_table"),
        df=make_df(
            sql_client=sql_client,
            columns=["account_id", "customer_id", "extra_dim", "ds_partitioned"],
            time_columns={"ds_partitioned"},
            data=bridge_data,
        ),
    )

    account_month_txns = [
        # While datetime columns can be used here, somewhere there is a conversion to a string when testing
        # with sqlite, so when you read the datetime column, it comes out as a string.
        # Columns for reference
        # ["account_id", DEFAULT_DS, "account_month", "txn_count", "txns_value", "ds_partitioned"]
        ("a0", "2020-01-01", "JANUARY", 3, 300, "2020-01-01"),
        ("a0", "2020-02-01", "FEBRUARY", 4, 400, "2020-02-01"),
        ("a0", "2020-03-01", "MARCH", 5, 1000, "2020-03-01"),
        ("a0", "2020-04-01", "APRIL", 4, 600, "2020-04-01"),
        ("a1", "2020-01-01", "JANUARY", 2, 200, "2020-01-01"),
        ("a1", "2020-02-01", "FEBRUARY", 5, 250, "2020-02-01"),
        ("a1", "2020-03-01", "MARCH", 6, 360, "2020-03-01"),
        ("a1", "2020-04-01", "APRIL", 10, 1000, "2020-04-01"),
        ("a2", "2020-01-01", "JANUARY", 5, 500, "2020-01-01"),
    ]
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="account_month_txns"),
        df=make_df(
            sql_client=sql_client,
            columns=["account_id", DEFAULT_DS, "account_month", "txn_count", "txns_value", "ds_partitioned"],
            time_columns={DEFAULT_DS, "ds_partitioned"},
            data=account_month_txns,
        ),
    )
    return True


@pytest.fixture(scope="session")
def create_message_source_tables(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> bool:
    """Creates tables with example source data for testing."""
    schema = mf_test_session_state.mf_source_schema

    data = [
        ("t0", "u0", "m0", "2020-01-01"),
        ("t0", "u0", "m1", "2020-01-01"),
        ("t0", "u1", "m2", "2020-01-01"),
        ("t0", "u1", "m3", "2020-01-01"),
        ("t0", "u1", "m4", "2020-01-01"),
        ("t1", "u0", "m5", "2020-01-02"),
        ("t1", "u0", "m7", "2020-01-02"),
        ("t1", "u0", "m8", "2020-01-02"),
        ("t1", "u1", "m9", "2020-01-02"),
        ("t1", "u1", "m10", "2020-01-02"),
        ("t1", "u1", "m11", "2020-01-02"),
        ("t1", "u1", "m12", "2020-01-03"),
        ("t1", "u2", "m13", "2020-01-03"),
        ("t1", "u3", "m14", "2020-01-03"),
    ]
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="fct_messages"),
        df=make_df(
            sql_client=sql_client,
            columns=["team_id", "user_id", "message_id", DEFAULT_DS],
            time_columns={DEFAULT_DS},
            data=data,
        ),
    )
    user_data = [
        ("t0", "u0", "burkina faso", "2020-01-01", "u0_2"),
        ("t0", "u1", "vanuatu", "2020-01-01", "u1_2"),
        ("t1", "u0", "merica", "2020-01-01", "u0_2"),
        ("t1", "u1", "merica", "2020-01-01", "u1_2"),
        ("t1", "u2", "burkina faso", "2020-01-01", "u2_2"),
        ("t1", "u3", "vanuata", "2020-01-01", "u3_2"),
    ]
    # Like fct_bookings, but using dt instead of ds to verify that a primary time dimension named something other than
    # ds can be used.
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="fct_users"),
        df=make_df(
            sql_client=sql_client,
            columns=["team_id", "id", "country", "created_at", "ident_2"],
            time_columns={"created_at"},
            data=user_data,
        ),
    )

    users_more_data = [
        ("u0", "thorium", "u0_2"),
        ("u1", "tellurium", "u1_2"),
        ("u0", "osmium", "u0_2"),
        ("u1", "tecnetium", "u1_2"),
        ("u2", "einsteinium", "u2_2"),
        ("u3", "rubidium", "u3_2"),
    ]
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="fct_users_more"),
        df=pd.DataFrame(
            columns=["id", "user_element", "ident_2"],
            data=users_more_data,
        ),
    )
    return True


@pytest.fixture(scope="session")
def create_extended_date_model_tables(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> bool:
    """Creates fct_bookings_extended / dim_listings_extended with a year's worth of dates for granularity testing."""
    schema = mf_test_session_state.mf_source_schema

    # Create fct_bookings_extended that has rows for every date in 2020.
    # Also create dim_listings_extended to match.
    start_date = datetime.datetime(2020, 1, 1)
    end_date = datetime.datetime(2020, 12, 31)

    current_date = start_date
    fct_bookings_extended_data = []
    dim_listings_extended_data = []
    booking_id = 0
    listing_id = 0
    while current_date <= end_date:
        ds_as_str = current_date.strftime(ISO8601_PYTHON_FORMAT)
        fct_bookings_extended_data.append((booking_id, 1, listing_id, booking_id % 2 == 0, ds_as_str))
        dim_listings_extended_data.append((listing_id, ds_as_str))

        listing_id += 1
        # Have 10 different listing IDs so that we get different distinct values when we set the granularity
        # to daily / weekly / monthly
        listing_id = listing_id % 10
        current_date = current_date + datetime.timedelta(days=1)

        booking_id += 1

    sql_table = SqlTable(schema_name=schema, table_name="fct_bookings_extended")
    sql_client.drop_table(sql_table)

    sql_client.create_table_from_dataframe(
        sql_table=sql_table,
        df=make_df(
            sql_client=sql_client,
            columns=["booking_id", "booking", "listing_id", "is_instant", DEFAULT_DS],
            time_columns={DEFAULT_DS},
            data=fct_bookings_extended_data,
        ),
    )

    sql_table = SqlTable(schema_name=schema, table_name="dim_listings_extended")
    sql_client.drop_table(sql_table)

    sql_client.create_table_from_dataframe(
        sql_table=sql_table,
        df=make_df(
            sql_client=sql_client,
            columns=["listing_id", "listing_creation_ds"],
            time_columns={"listing_creation_ds"},
            data=dim_listings_extended_data,
        ),
    )

    # Create fct_bookings_extended_monthly with data at a monthly granularity.
    # Example rows:
    # ["bookings_monthly", "is_instant", "ds"]
    # 10                   True          2020-01-01
    # 5                    False         2020-01-01
    # 10                   True          2020-02-01
    # 5                    False         2020-02-01
    current_date = pd.Timestamp(start_date)
    fct_bookings_extended_monthly_data = []
    while current_date <= end_date:
        fct_bookings_extended_monthly_data.append((10, True, current_date.strftime(ISO8601_PYTHON_FORMAT)))
        fct_bookings_extended_monthly_data.append((5, False, current_date.strftime(ISO8601_PYTHON_FORMAT)))

        current_date += TimeGranularity.MONTH.period_begin_offset

    sql_table = SqlTable(schema_name=schema, table_name="fct_bookings_extended_monthly")
    sql_client.drop_table(sql_table=sql_table)
    sql_client.create_table_from_dataframe(
        sql_table=sql_table,
        df=make_df(
            sql_client=sql_client,
            columns=["bookings_monthly", "is_instant", DEFAULT_DS],
            time_columns={DEFAULT_DS},
            data=fct_bookings_extended_monthly_data,
        ),
    )

    return True


@pytest.fixture(scope="session")
def create_data_warehouse_validation_model_tables(
    mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient
) -> bool:
    """Creates tables with example source data for testing."""
    schema = mf_test_session_state.mf_source_schema

    data = [
        (True, "2022-06-01"),
    ]
    create_table(
        sql_client=sql_client,
        sql_table=SqlTable(schema_name=schema, table_name="fct_animals"),
        df=make_df(
            sql_client=sql_client,
            columns=["is_dog", DEFAULT_DS],
            time_columns={DEFAULT_DS},
            data=data,
        ),
    )

    return True



from abc import ABC
from dataclasses import dataclass
from typing import Optional, Any

from metricflow.dataclass_serialization import SerializableDataclass
from metricflow.object_utils import assert_exactly_one_arg_set


class ColumnCorrelationKey(ABC):
    """Interface for a key object that is used to correlate columns between instance sets."""

    pass


@dataclass(frozen=True)
class SingleColumnCorrelationKey(ColumnCorrelationKey, SerializableDataclass):
    """Key to use when there's only 1 column association in an instance."""

    # Pydantic throws an error during serialization if a dataclass has no fields.
    __PYDANTIC_BUG_WORKAROUND: bool = True

    def __eq__(self, other: Any) -> bool:  # type: ignore[misc] # noqa: D
        return isinstance(other, SingleColumnCorrelationKey)

    def __hash__(self) -> int:  # noqa: D
        return hash(self.__class__.__name__)


@dataclass(frozen=True)
class CompositeColumnCorrelationKey(ColumnCorrelationKey, SerializableDataclass):
    """Key to use when there are multiple column associations in an instance"""

    sub_identifier: str


@dataclass(frozen=True)
class ColumnAssociation(SerializableDataclass):
    """Used to describe how an instance is associated with columns in table / SQL query.

    Generally there will be a 1:1 mapping, but for composite identifiers, it can map to multiple columns. For that case,
    this can be subclassed to add more context.
    """

    column_name: str
    # When an instance is passed from one dataflow node to another, we need to know how the columns from the input
    # corresponds to the columns from the output. Equality of this key is used to determine that relationship.
    # This could be made to be in a dictionary instead, but having it here means that it doesn't need to be hashable.
    single_column_correlation_key: Optional[SingleColumnCorrelationKey] = None
    composite_column_correlation_key: Optional[CompositeColumnCorrelationKey] = None

    def __post_init__(self) -> None:  # noqa: D
        assert_exactly_one_arg_set(
            single_column_correlation_key=self.single_column_correlation_key,
            composite_column_correlation_key=self.composite_column_correlation_key,
        )

    @property
    def column_correlation_key(self) -> ColumnCorrelationKey:  # noqa: D
        if self.single_column_correlation_key:
            return self.single_column_correlation_key
        elif self.composite_column_correlation_key:
            return self.composite_column_correlation_key

        raise RuntimeError("single_column_correlation_key or composite_column_correlation_key should be set")



from _pytest.fixtures import FixtureRequest

from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.model.semantic_model import SemanticModel
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.dataflow_to_execution import DataflowToExecutionPlanConverter
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs import (
    MetricFlowQuerySpec,
    MetricSpec,
    DimensionSpec,
    TimeDimensionSpec,
    IdentifierReference,
)
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.plan_utils import assert_execution_plan_text_equal


def make_execution_plan_converter(  # noqa: D
    semantic_model: SemanticModel,
    sql_client: SqlClient,
    time_spine_source: TimeSpineSource,
) -> DataflowToExecutionPlanConverter:
    return DataflowToExecutionPlanConverter[DataSourceDataSet](
        sql_plan_converter=DataflowToSqlQueryPlanConverter[DataSourceDataSet](
            column_association_resolver=DefaultColumnAssociationResolver(semantic_model),
            semantic_model=semantic_model,
            time_spine_source=time_spine_source,
        ),
        sql_plan_renderer=DefaultSqlQueryPlanRenderer(),
        sql_client=sql_client,
    )


def test_joined_plan(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    dataflow_plan_builder: DataflowPlanBuilder[DataSourceDataSet],
    simple_semantic_model: SemanticModel,
    sql_client: SqlClient,
    time_spine_source: TimeSpineSource,
) -> None:
    dataflow_plan = dataflow_plan_builder.build_plan(
        MetricFlowQuerySpec(
            metric_specs=(MetricSpec(element_name="bookings"),),
            dimension_specs=(
                DimensionSpec(
                    element_name="is_instant",
                    identifier_links=(),
                ),
                DimensionSpec(
                    element_name="country_latest",
                    identifier_links=(IdentifierReference("listing"),),
                ),
            ),
        )
    )

    to_execution_plan_converter = make_execution_plan_converter(simple_semantic_model, sql_client, time_spine_source)

    execution_plan = to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)

    assert_execution_plan_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_client=sql_client,
        execution_plan=execution_plan,
    )


def test_small_combined_metrics_plan(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    dataflow_plan_builder: DataflowPlanBuilder[DataSourceDataSet],
    simple_semantic_model: SemanticModel,
    time_spine_source: TimeSpineSource,
) -> None:
    dataflow_plan = dataflow_plan_builder.build_plan(
        MetricFlowQuerySpec(
            metric_specs=(
                MetricSpec(element_name="bookings"),
                MetricSpec(element_name="booking_value"),
            ),
            dimension_specs=(
                DimensionSpec(
                    element_name="is_instant",
                    identifier_links=(),
                ),
            ),
        )
    )

    to_execution_plan_converter = make_execution_plan_converter(
        semantic_model=simple_semantic_model,
        sql_client=sql_client,
        time_spine_source=time_spine_source,
    )
    execution_plan = to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)

    assert_execution_plan_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_client=sql_client,
        execution_plan=execution_plan,
    )


def test_combined_metrics_plan(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    dataflow_plan_builder: DataflowPlanBuilder[DataSourceDataSet],
    simple_semantic_model: SemanticModel,
    time_spine_source: TimeSpineSource,
) -> None:
    dataflow_plan = dataflow_plan_builder.build_plan(
        MetricFlowQuerySpec(
            metric_specs=(
                MetricSpec(element_name="bookings"),
                MetricSpec(element_name="instant_bookings"),
                MetricSpec(element_name="booking_value"),
            ),
            dimension_specs=(
                DimensionSpec(
                    element_name="is_instant",
                    identifier_links=(),
                ),
            ),
            time_dimension_specs=(TimeDimensionSpec(element_name="ds", identifier_links=()),),
        )
    )

    to_execution_plan_converter = make_execution_plan_converter(
        semantic_model=simple_semantic_model,
        sql_client=sql_client,
        time_spine_source=time_spine_source,
    )
    execution_plan = to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)

    assert_execution_plan_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_client=sql_client,
        execution_plan=execution_plan,
    )


def test_multihop_joined_plan(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    multihop_dataflow_plan_builder: DataflowPlanBuilder[DataSourceDataSet],
    multi_hop_join_semantic_model: SemanticModel,
    sql_client: SqlClient,
    time_spine_source: TimeSpineSource,
) -> None:
    """Tests a plan getting a measure and a joined dimension."""
    dataflow_plan = multihop_dataflow_plan_builder.build_plan(
        MetricFlowQuerySpec(
            metric_specs=(MetricSpec(element_name="txn_count"),),
            dimension_specs=(
                DimensionSpec(
                    element_name="customer_name",
                    identifier_links=(
                        IdentifierReference(element_name="account_id"),
                        IdentifierReference(element_name="customer_id"),
                    ),
                ),
            ),
        )
    )

    to_execution_plan_converter = DataflowToExecutionPlanConverter(
        sql_plan_converter=DataflowToSqlQueryPlanConverter[DataSourceDataSet](
            column_association_resolver=DefaultColumnAssociationResolver(multi_hop_join_semantic_model),
            semantic_model=multi_hop_join_semantic_model,
            time_spine_source=time_spine_source,
        ),
        sql_plan_renderer=DefaultSqlQueryPlanRenderer(),
        sql_client=sql_client,
    )

    execution_plan = to_execution_plan_converter.convert_to_execution_plan(dataflow_plan)

    assert_execution_plan_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_client=sql_client,
        execution_plan=execution_plan,
    )



import difflib
import logging
import os
import re
import webbrowser
from typing import TypeVar, Optional, Tuple, List, Callable

from _pytest.fixtures import FixtureRequest

from metricflow.dag.mf_dag import MetricFlowDag
from metricflow.dataflow.dataflow_plan import DataflowPlan
from metricflow.dataflow.dataflow_plan_to_text import dataflow_plan_as_text
from metricflow.execution.execution_plan import ExecutionPlan
from metricflow.execution.execution_plan_to_text import execution_plan_to_text
from metricflow.protocols.sql_client import SqlClient
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState

logger = logging.getLogger(__name__)


# In plan outputs, replace strings that vary from run to run with this so that comparisons can be made
# consistently.
PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS = "*"


def make_schema_replacement_function(system_schema: str, source_schema: str) -> Callable[[str], str]:
    """Generates a function to replace schema names in test outputs."""
    # The schema of the warehouse used in tests changes from run to run, so don't compare those.
    def replacement_function(text: str) -> str:  # noqa: D
        # Replace with a string of the same length so that indents are preserved.
        text = text.replace(source_schema, PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS * len(source_schema))
        # Same with the MetricFlow system schema.
        return text.replace(system_schema, PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS * len(system_schema))

    return replacement_function


def replace_dataset_id_hash(text: str) -> str:
    """Replaces data set ID hashes for primed data sources.

    The data set ID hash changes from run to run because it's based on the DW schema in the data source, which changes
    run to run.
    """
    pattern = re.compile(r"'[a-zA-Z0-9_]+__[a-zA-Z0-9_]+__(?P<hash>[a-zA-Z0-9_]+)'")
    while True:
        match = pattern.search(text)
        if match:
            data_set_id_hash = match.group("hash")
            # Replace with the same length to preserve indents
            text = text.replace(data_set_id_hash, PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS * len(data_set_id_hash))
        else:
            break
    return text


def snapshot_path_prefix(
    request: FixtureRequest,
    snapshot_group: str,
    snapshot_id: str,
    additional_sub_directories: Tuple[str, ...] = (),
) -> str:
    """Returns a path prefix that can be used to build filenames for files associated with the snapshot.

    The snapshot prefix is generated from the name of the test file, the name of the test, name of the snapshot class,
    and the name of the snapshot.

    e.g.
    .../snapshots/test_file.py/DataflowPlan/test_name__plan1

    which can be used to construct paths like

    .../snapshots/test_file.py/DataflowPlan/test_name__plan1.xml
    .../snapshots/test_file.py/DataflowPlan/test_name__plan1.svg
    """
    test_name = request.node.name
    snapshot_file_name = test_name + "__" + snapshot_id
    path_items: List[str] = []

    test_file_path_items = os.path.normpath(request.node.fspath).split(os.sep)
    test_file_name = test_file_path_items[-1]
    # Default to where this is defined, but use more appropriate directories if found.
    test_directory_root_index = -1
    for i, path_item in enumerate(test_file_path_items):
        if path_item in ("test", "metricflow", "metricflow_extensions"):
            test_directory_root_index = i + 1

    path_to_store_snapshots = os.sep.join(test_file_path_items[:test_directory_root_index])
    path_items.extend([path_to_store_snapshots, "snapshots", test_file_name, snapshot_group])

    if additional_sub_directories:
        path_items.extend(additional_sub_directories)
    path_items.append(snapshot_file_name)

    return os.path.abspath(os.path.join(*path_items))


PlanT = TypeVar("PlanT", bound=MetricFlowDag)


def _exclude_lines_matching_regex(file_contents: str, exclude_line_regex: str) -> str:
    """Removes lines from file contents if the line matches exclude_regex."""
    compiled_regex = re.compile(exclude_line_regex)
    return "\n".join([line for line in file_contents.split("\n") if not compiled_regex.match(line)])


def assert_plan_snapshot_text_equal(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    plan: PlanT,
    plan_snapshot_text: str,
    plan_snapshot_file_extension: str = ".xml",
    exclude_line_regex: Optional[str] = None,
    incomparable_strings_replacement_function: Callable[[str], str] = None,
    additional_sub_directories_for_snapshots: Tuple[str, ...] = (),
) -> None:
    """Checks if the given plan text is equal to the one that's saved for comparison.

    * The location of the file is automatically generated based on the test and the plan's ID.
    * This may create a new saved plan file or overwrite the existing one, depending on the configuration.
    * replace_incomparable_strings is used to replace strings in the plan text before comparison. Useful for making
      plans consistent when there are strings that vary between runs and shouldn't be compared.
    * additional_sub_directories_for_snapshots is used to specify additional sub-directories (in the automatically
      generated directory) where plan outputs should reside.
    """
    assert_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        group_id=plan.__class__.__name__,
        snapshot_id=plan.dag_id,
        snapshot_text=plan_snapshot_text,
        snapshot_file_extension=plan_snapshot_file_extension,
        exclude_line_regex=exclude_line_regex,
        incomparable_strings_replacement_function=incomparable_strings_replacement_function,
        additional_sub_directories_for_snapshots=additional_sub_directories_for_snapshots,
    )


def assert_snapshot_text_equal(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    group_id: str,
    snapshot_id: str,
    snapshot_text: str,
    snapshot_file_extension: str,
    exclude_line_regex: Optional[str] = None,
    incomparable_strings_replacement_function: Callable[[str], str] = None,
    additional_sub_directories_for_snapshots: Tuple[str, ...] = (),
) -> None:
    """Similar to assert_plan_snapshot_text_equal(), but with more controls on how the snapshot paths are generated."""
    file_path = (
        snapshot_path_prefix(
            request=request,
            snapshot_group=group_id,
            snapshot_id=snapshot_id,
            additional_sub_directories=additional_sub_directories_for_snapshots,
        )
        + snapshot_file_extension
    )

    if incomparable_strings_replacement_function:
        snapshot_text = incomparable_strings_replacement_function(snapshot_text)

    # If we are in overwrite mode, make a new plan:
    if mf_test_session_state.overwrite_snapshots:
        # Create parent directory for the plan text files.
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w") as snapshot_text_file:
            snapshot_text_file.write(snapshot_text)
            # Add a new line at the end of the file so that PRSs don't show the "no newline" symbol on Github.
            snapshot_text_file.write("\n")

    # Throw an exception if the plan is not there.
    if not os.path.exists(file_path):
        raise FileNotFoundError(
            f"Could not find snapshot file at path {file_path}. Re-run with --overwrite-snapshots and check git status "
            "to see what's new."
        )

    if mf_test_session_state.display_plans:
        if not mf_test_session_state.overwrite_snapshots:
            logger.warning(f"Not overwriting snapshots, so displaying existing snapshot at {file_path}")

        if mf_test_session_state.plans_displayed >= mf_test_session_state.max_plans_displayed:
            raise RuntimeError(
                f"Can't display snapshot - hit limit of "
                f"{mf_test_session_state.max_plans_displayed} "
                f"plans displayed."
            )
        webbrowser.open("file://" + file_path)
        mf_test_session_state.plans_displayed += 1

    # Read the existing plan from the file and compare with the actual plan
    with open(file_path, "r") as snapshot_text_file:
        # Remove the newline that was added from above.
        expected_snapshot_text = snapshot_text_file.read().rstrip()

        if exclude_line_regex:
            # Filter out lines that should be ignored.
            expected_snapshot_text = _exclude_lines_matching_regex(
                file_contents=expected_snapshot_text, exclude_line_regex=exclude_line_regex
            )
            snapshot_text = _exclude_lines_matching_regex(
                file_contents=snapshot_text, exclude_line_regex=exclude_line_regex
            )
        # pytest should show a detailed diff with "assert actual_modified == expected_modified", but it's not, so doing
        # this instead.
        if snapshot_text != expected_snapshot_text:
            differ = difflib.Differ()
            diff = differ.compare(expected_snapshot_text.splitlines(), snapshot_text.splitlines())
            assert False, f"Snapshot from {file_path} does not match. Diff from expected to actual:\n" + "\n".join(diff)


def assert_execution_plan_text_equal(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    execution_plan: ExecutionPlan,
) -> None:
    assert_plan_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan=execution_plan,
        plan_snapshot_text=execution_plan_to_text(execution_plan),
        incomparable_strings_replacement_function=make_schema_replacement_function(
            system_schema=mf_test_session_state.mf_system_schema,
            source_schema=mf_test_session_state.mf_source_schema,
        ),
        additional_sub_directories_for_snapshots=(sql_client.__class__.__name__,),
    )


def assert_dataflow_plan_text_equal(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    dataflow_plan: DataflowPlan,
    sql_client: SqlClient,
) -> None:
    assert_plan_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan=dataflow_plan,
        plan_snapshot_text=dataflow_plan_as_text(dataflow_plan),
        incomparable_strings_replacement_function=replace_dataset_id_hash,
        additional_sub_directories_for_snapshots=(sql_client.__class__.__name__,),
    )



import logging

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.transformations.transform_rule import ModelTransformRule

logger = logging.getLogger(__name__)


class LowerCaseNamesRule(ModelTransformRule):
    """Lowercases the names of both top level objects and data source elements in a model"""

    @staticmethod
    def transform_model(model: UserConfiguredModel) -> UserConfiguredModel:  # noqa: D
        LowerCaseNamesRule._lowercase_top_level_objects(model)
        for data_source in model.data_sources:
            LowerCaseNamesRule._lowercase_data_source_elements(data_source)

        return model

    @staticmethod
    def _lowercase_data_source_elements(data_source: DataSource) -> None:
        """Lowercases the names of data source elements."""
        if data_source.measures:
            for measure in data_source.measures:
                measure.name = measure.name.lower()
        if data_source.identifiers:
            for identifier in data_source.identifiers:
                identifier.name = identifier.name.lower()
        if data_source.dimensions:
            for dimension in data_source.dimensions:
                dimension.name = dimension.name.lower()

    @staticmethod
    def _lowercase_top_level_objects(model: UserConfiguredModel) -> None:
        """Lowercases the names of model objects"""
        if model.data_sources:
            for data_source in model.data_sources:
                data_source.name = data_source.name.lower()
        if model.materializations:
            for materialization in model.materializations:
                materialization.name = materialization.name.lower()



from __future__ import annotations

import logging
from collections import defaultdict, OrderedDict
from dataclasses import dataclass
from typing import Tuple, List, Dict, Sequence, Set, Optional

import pandas as pd

from metricflow.constraints.time_constraint import TimeRangeConstraint
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.dataflow_plan import BaseOutput
from metricflow.dataset.data_source_adapter import DataSourceDataSet
from metricflow.instances import MetricModelReference
from metricflow.model.objects.metric import MetricType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.semantic_model import SemanticModel
from metricflow.query.query_exceptions import InvalidQueryException
from metricflow.references import TimeDimensionReference, MeasureReference, MetricReference, IdentifierReference
from metricflow.specs import (
    TimeDimensionSpec,
    DEFAULT_TIME_GRANULARITY,
)
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class PartialTimeDimensionSpec:
    """Similar to TimeDimensionSpec, but with an unknown time granularity.

    This is used to represent a time dimension spec from the user before the granularity is figured out.
    """

    element_name: str
    identifier_links: Tuple[IdentifierReference, ...]


@dataclass(frozen=True)
class LocalTimeDimensionGranularityKey:
    """Used to associate a measure and a local time dimension without being specific about a granularity."""

    measure_reference: MeasureReference
    local_time_dimension_reference: TimeDimensionReference


class TimeGranularitySolver:
    """Figures out time granularities for metrics.

    1. Given the list of metrics, what's the smallest time granularity that's common across those metrics?
    2. Given the list of metrics and the dimensions that the user specified, does the primary time dimension need to be
    converted into another granularity? e.g. if the metrics have different time granularities and smallest time
    granularity for the given metrics is MONTH and the user specified ds, then we need to convert ds to ds__month.
    3. Also checks to see if the time granularity expressed in dimensions is valid. e.g. if a metric has a time
    granularity of MONTH, throw an exception if ds__day is in the requested dimensions.
    """

    def __init__(  # noqa: D
        self,
        semantic_model: SemanticModel,
        source_nodes: Sequence[BaseOutput[DataSourceDataSet]],
        node_output_resolver: DataflowPlanNodeOutputDataSetResolver[DataSourceDataSet],
    ) -> None:
        self._semantic_model = semantic_model
        self._metric_reference_to_measure_reference = TimeGranularitySolver._measures_for_metric(
            self._semantic_model.user_configured_model
        )

        self._local_time_dimension_granularities: Dict[
            LocalTimeDimensionGranularityKey, Set[TimeGranularity]
        ] = defaultdict(set)

        for source_node in source_nodes:
            output_data_set = node_output_resolver.get_output_data_set(source_node)
            for time_dimension_instance in output_data_set.instance_set.time_dimension_instances:
                time_dimension_spec = time_dimension_instance.spec
                if len(time_dimension_spec.identifier_links) == 0:
                    for measure_instance in output_data_set.instance_set.measure_instances:
                        self._local_time_dimension_granularities[
                            LocalTimeDimensionGranularityKey(
                                measure_reference=measure_instance.spec.as_reference,
                                local_time_dimension_reference=time_dimension_spec.reference,
                            )
                        ].add(time_dimension_spec.time_granularity)

    @staticmethod
    def _measures_for_metric(model: UserConfiguredModel) -> Dict[MetricModelReference, List[MeasureReference]]:
        """Given a model, return a dict that maps the name of the metric to the names of the measures used"""
        metric_reference_to_measure_references: Dict[MetricModelReference, List[MeasureReference]] = {}
        for metric in model.metrics:
            metric_reference_to_measure_references[MetricModelReference(metric_name=metric.name)] = [
                MeasureReference(element_name=measure.element_name) for measure in metric.measure_references
            ]

        return metric_reference_to_measure_references

    def local_dimension_granularity_range(
        self, metric_references: Sequence[MetricReference], local_time_dimension_reference: TimeDimensionReference
    ) -> Tuple[TimeGranularity, TimeGranularity]:
        """Return the ranges of time granularities for a local dimension associated with the measures in metrics.

        For example, let's say we're querying for two metrics that are based on 'bookings' and 'bookings_monthly'
        respectively.

        The 'bookings' measure defined is in the 'fct_bookings' data source. 'fct_bookings' has a local time dimension
        named 'ds' with granularity DAY.

        The 'monthly_bookings' measure is in defined in the 'fct_bookings_monthly' data source. 'fct_bookings_monthly'
        has a local time dimension named 'ds' with granularity DAY.

        Then this would return [DAY, MONTH].
        """
        all_time_granularities = set()

        for metric_reference in metric_references:
            for measure_reference in self._metric_reference_to_measure_reference[
                MetricModelReference(metric_name=metric_reference.element_name)
            ]:
                key = LocalTimeDimensionGranularityKey(
                    measure_reference=measure_reference,
                    local_time_dimension_reference=local_time_dimension_reference,
                )
                valid_time_granularities_for_measure = self._local_time_dimension_granularities[key]

                if len(valid_time_granularities_for_measure) == 0:
                    raise InvalidQueryException(
                        f"Local dimension {local_time_dimension_reference} does not exist for measure "
                        f"'{measure_reference}' of metric '{metric_reference}"
                    )

                all_time_granularities.add(min(valid_time_granularities_for_measure))

        return min(all_time_granularities), max(all_time_granularities)

    def validate_time_granularity(
        self, metric_references: Sequence[MetricReference], time_dimension_specs: Sequence[TimeDimensionSpec]
    ) -> None:
        """Check that the granularity specified for time dimensions is valid with respect to the metrics.

        e.g. throw an error if "ds__week" is specified for a metric with a time granularity of MONTH.
        """
        for time_dimension_spec in time_dimension_specs:
            # Validate local time dimensions.
            if time_dimension_spec.identifier_links == ():
                _, min_granularity_for_querying = self.local_dimension_granularity_range(
                    metric_references=metric_references,
                    local_time_dimension_reference=time_dimension_spec.reference,
                )
                if time_dimension_spec.time_granularity < min_granularity_for_querying:
                    raise RequestTimeGranularityException(
                        f"The minimum time granularity for querying metrics {metric_references} is "
                        f"{min_granularity_for_querying}. Got {time_dimension_spec}"
                    )

                # If there is a cumulative metric, granularity changes aren't supported.
                for metric_reference in metric_references:
                    metric = self._semantic_model.metric_semantics.get_metric(metric_reference)
                    if (
                        metric.type == MetricType.CUMULATIVE
                        and time_dimension_spec.time_granularity != min_granularity_for_querying
                    ):
                        raise RequestTimeGranularityException(
                            f"For querying cumulative metric '{metric_reference.element_name}', the granularity of "
                            f"'{time_dimension_spec.qualified_name}' must be {min_granularity_for_querying.name}"
                        )

        # TODO: Validate non-local time dimension granularities here instead of during plan building.

    def resolve_granularity_for_partial_time_dimension_specs(
        self,
        metric_references: Sequence[MetricReference],
        partial_time_dimension_specs: Sequence[PartialTimeDimensionSpec],
        metric_time_dimension_reference: TimeDimensionReference,
        time_granularity: Optional[TimeGranularity] = None,
    ) -> Dict[PartialTimeDimensionSpec, TimeDimensionSpec]:
        """Figure out the lowest granularity possible for the partially specified time dimension specs.

        Returns a dictionary that maps how the partial time dimension spec should be turned into a time dimension spec.
        """
        replacement_dict: OrderedDict[PartialTimeDimensionSpec, TimeDimensionSpec] = OrderedDict()
        for partial_time_dimension_spec in partial_time_dimension_specs:
            # Handle local time dimensions
            if partial_time_dimension_spec.identifier_links == ():
                _, min_time_granularity_for_querying = self.local_dimension_granularity_range(
                    metric_references=metric_references,
                    local_time_dimension_reference=TimeDimensionReference(
                        element_name=partial_time_dimension_spec.element_name
                    ),
                )

                if (
                    partial_time_dimension_spec.element_name == metric_time_dimension_reference.element_name
                    and time_granularity
                ):
                    if time_granularity < min_time_granularity_for_querying:
                        raise RequestTimeGranularityException(
                            f"Can't use time granularity for time dimension '{metric_time_dimension_reference} since "
                            f"the minimum granularity is {min_time_granularity_for_querying}"
                        )
                    replacement_dict[partial_time_dimension_spec] = TimeDimensionSpec(
                        element_name=partial_time_dimension_spec.element_name,
                        identifier_links=(),
                        time_granularity=time_granularity,
                    )
                else:
                    replacement_dict[partial_time_dimension_spec] = TimeDimensionSpec(
                        element_name=partial_time_dimension_spec.element_name,
                        identifier_links=(),
                        time_granularity=min_time_granularity_for_querying,
                    )
            # Handle joined time dimensions.
            else:
                # TODO: For joined time dimensions, also compute the minimum granularity for querying.
                replacement_dict[partial_time_dimension_spec] = TimeDimensionSpec(
                    element_name=partial_time_dimension_spec.element_name,
                    identifier_links=partial_time_dimension_spec.identifier_links,
                    time_granularity=DEFAULT_TIME_GRANULARITY,
                )
        return replacement_dict

    def adjust_time_range_to_granularity(
        self, time_range_constraint: TimeRangeConstraint, time_granularity: TimeGranularity
    ) -> TimeRangeConstraint:
        """Change the time range so that the ends are at the ends of the appropriate time granularity windows.

        e.g. [2020-01-15, 2020-2-15] with MONTH granularity -> [2020-01-01, 2020-02-29]
        """
        constraint_start = time_range_constraint.start_time
        constraint_end = time_range_constraint.end_time

        start_ts = pd.Timestamp(time_range_constraint.start_time)
        if not time_granularity.is_period_start(start_ts):
            constraint_start = time_granularity.adjust_to_start_of_period(start_ts).to_pydatetime()

        end_ts = pd.Timestamp(time_range_constraint.end_time)
        if not time_granularity.is_period_end(end_ts):
            constraint_end = time_granularity.adjust_to_end_of_period(end_ts).to_pydatetime()

        if constraint_start < TimeRangeConstraint.ALL_TIME_BEGIN():
            constraint_start = TimeRangeConstraint.ALL_TIME_BEGIN()
        if constraint_end > TimeRangeConstraint.ALL_TIME_END():
            constraint_end = TimeRangeConstraint.ALL_TIME_END()

        return TimeRangeConstraint(start_time=constraint_start, end_time=constraint_end)


class RequestTimeGranularityException(Exception):
    """Raised when a query is requesting a time granularity that's not possible for the given metrics."""

    pass



from typing import Sequence

import pytest

from metricflow.specs import (
    DimensionSpec,
    IdentifierSpec,
    InstanceSpec,
    LinkableInstanceSpec,
    TimeDimensionSpec,
    InstanceSpecSet,
    MetricSpec,
    MeasureSpec,
    LinklessIdentifierSpec,
    IdentifierReference,
)
from metricflow.time.time_granularity import TimeGranularity


@pytest.fixture
def dimension_spec() -> DimensionSpec:  # noqa: D
    return DimensionSpec(
        element_name="platform",
        identifier_links=(
            IdentifierReference(element_name="user_id"),
            IdentifierReference(element_name="device_id"),
        ),
    )


@pytest.fixture
def time_dimension_spec() -> TimeDimensionSpec:  # noqa: D
    return TimeDimensionSpec(
        element_name="signup_ts",
        identifier_links=(IdentifierReference(element_name="user_id"),),
        time_granularity=TimeGranularity.DAY,
    )


@pytest.fixture
def identifier_spec() -> IdentifierSpec:  # noqa: D
    return IdentifierSpec(
        element_name="user_id",
        identifier_links=(IdentifierReference(element_name="listing_id"),),
    )


def test_merge_specs(dimension_spec: DimensionSpec, identifier_spec: IdentifierSpec) -> None:
    """Tests InstanceSpec.merge()"""
    assert InstanceSpec.merge([dimension_spec], [identifier_spec]) == [dimension_spec, identifier_spec]


def test_dimension_without_first_identifier_link(dimension_spec: DimensionSpec) -> None:  # noqa: D
    assert dimension_spec.without_first_identifier_link() == DimensionSpec(
        element_name="platform", identifier_links=(IdentifierReference(element_name="device_id"),)
    )


def test_dimension_without_identifier_links(dimension_spec: DimensionSpec) -> None:  # noqa: D
    assert dimension_spec.without_identifier_links() == DimensionSpec(element_name="platform", identifier_links=())


def test_time_dimension_without_first_identifier_link(time_dimension_spec: TimeDimensionSpec) -> None:  # noqa: D
    assert time_dimension_spec.without_first_identifier_link() == TimeDimensionSpec(
        element_name="signup_ts",
        identifier_links=(),
        time_granularity=TimeGranularity.DAY,
    )


def test_time_dimension_without_identifier_links(time_dimension_spec: TimeDimensionSpec) -> None:  # noqa: D
    assert time_dimension_spec.without_identifier_links() == LinklessIdentifierSpec(
        element_name="signup_ts",
        identifier_links=(),
    )


def test_identifier_without_first_identifier_link(identifier_spec: IdentifierSpec) -> None:  # noqa: D
    assert identifier_spec.without_first_identifier_link() == IdentifierSpec(
        element_name="user_id",
        identifier_links=(),
    )


def test_identifier_without_identifier_links(identifier_spec: IdentifierSpec) -> None:  # noqa: D
    assert identifier_spec.without_identifier_links() == IdentifierSpec(
        element_name="user_id",
        identifier_links=(),
    )


def test_merge_linkable_specs(dimension_spec: DimensionSpec, identifier_spec: IdentifierSpec) -> None:  # noqa: D
    linkable_specs: Sequence[LinkableInstanceSpec] = [dimension_spec, identifier_spec]

    assert LinkableInstanceSpec.merge_linkable_specs([dimension_spec], [identifier_spec]) == linkable_specs


def test_qualified_name() -> None:  # noqa: D
    assert (
        DimensionSpec(element_name="country", identifier_links=(IdentifierReference("listing_id"),)).qualified_name
        == "listing_id__country"
    )


def test_merge_spec_set() -> None:  # noqa: D
    spec_set1 = InstanceSpecSet(metric_specs=(MetricSpec(element_name="bookings"),))
    spec_set2 = InstanceSpecSet(dimension_specs=(DimensionSpec(element_name="is_instant", identifier_links=()),))

    assert spec_set1.merge([spec_set2]) == InstanceSpecSet(
        metric_specs=(MetricSpec(element_name="bookings"),),
        dimension_specs=(DimensionSpec(element_name="is_instant", identifier_links=()),),
    )


@pytest.fixture
def spec_set() -> InstanceSpecSet:  # noqa: D
    return InstanceSpecSet(
        metric_specs=(MetricSpec(element_name="bookings"),),
        measure_specs=(
            MeasureSpec(
                element_name="bookings",
            ),
        ),
        dimension_specs=(DimensionSpec(element_name="is_instant", identifier_links=()),),
        time_dimension_specs=(
            TimeDimensionSpec(
                element_name="ds",
                identifier_links=(),
                time_granularity=TimeGranularity.DAY,
            ),
        ),
        identifier_specs=(
            IdentifierSpec(
                element_name="user_id",
                identifier_links=(IdentifierReference(element_name="listing_id"),),
            ),
        ),
    )


def test_spec_set_linkable_specs(spec_set: InstanceSpecSet) -> None:  # noqa: D
    assert set(spec_set.linkable_specs) == {
        DimensionSpec(element_name="is_instant", identifier_links=()),
        TimeDimensionSpec(
            element_name="ds",
            identifier_links=(),
            time_granularity=TimeGranularity.DAY,
        ),
        IdentifierSpec(
            element_name="user_id",
            identifier_links=(IdentifierReference(element_name="listing_id"),),
        ),
    }


def test_spec_set_all_specs(spec_set: InstanceSpecSet) -> None:  # noqa: D
    assert set(spec_set.all_specs) == {
        MetricSpec(element_name="bookings"),
        MeasureSpec(
            element_name="bookings",
        ),
        DimensionSpec(element_name="is_instant", identifier_links=()),
        TimeDimensionSpec(
            element_name="ds",
            identifier_links=(),
            time_granularity=TimeGranularity.DAY,
        ),
        IdentifierSpec(
            element_name="user_id",
            identifier_links=(IdentifierReference(element_name="listing_id"),),
        ),
    }


def test_linkless_identifier() -> None:  # noqa: D
    """Check that equals and hash works as expected for the LinklessIdentifierSpec / IdentifierSpec"""
    identifier_spec = IdentifierSpec(element_name="user_id", identifier_links=())
    linkless_identifier_spec = LinklessIdentifierSpec.from_element_name("user_id")

    # Check equality between the two.
    assert identifier_spec == identifier_spec
    assert linkless_identifier_spec == linkless_identifier_spec
    assert identifier_spec == linkless_identifier_spec

    # Check that they are treated equivalently in sets.
    set_with_identifier_spec = {identifier_spec}
    assert identifier_spec in set_with_identifier_spec
    assert linkless_identifier_spec in set_with_identifier_spec

    set_with_linkless_identifier_spec = {linkless_identifier_spec}
    assert identifier_spec in set_with_linkless_identifier_spec
    assert linkless_identifier_spec in set_with_linkless_identifier_spec

    set_with_identifier_spec.add(linkless_identifier_spec)
    assert len(set_with_identifier_spec) == 1
